# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import binascii
import socket
from collections import OrderedDict

from smbprotocol import Dialects
from smbprotocol.connection import Capabilities, SecurityMode
from smbprotocol.header import Commands
from smbprotocol.structure import (
    BytesField,
    EnumField,
    FlagField,
    IntField,
    ListField,
    Structure,
    StructureField,
    UuidField,
)


class CtlCode:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31 SMB2 IOCTL Request CtlCode
    The control code of the FSCTL_IOCTL method.
    """

    FSCTL_DFS_GET_REFERRALS = 0x00060194
    FSCTL_PIPE_PEEK = 0x0011400C
    FSCTL_PIPE_WAIT = 0x00110018
    FSCTL_PIPE_TRANSCEIVE = 0x0011C017
    FSCTL_SRV_COPYCHUNK = 0x001440F2
    FSCTL_SRV_ENUMERATE_SNAPSHOTS = 0x00144064
    FSCTL_SRV_REQUEST_RESUME_KEY = 0x00140078
    FSCTL_SRV_READ_HASH = 0x001441BB
    FSCTL_SRV_COPYCHUNK_WRITE = 0x001480F2
    FSCTL_LMR_REQUEST_RESILIENCY = 0x001401D4
    FSCTL_QUERY_NETWORK_INTERFACE_INFO = 0x001401FC
    FSCTL_SET_REPARSE_POINT = 0x000900A4
    FSCTL_GET_REPARSE_POINT = 0x000900A8
    FSCTL_DELETE_REPARSE_POINT = 0x000900AC
    FSCTL_CREATE_OR_GET_OBJECT_ID = 0x000900C0
    FSCTL_SET_SPARSE = 0x000900C4
    FSCTL_QUERY_ALLOCATED_RANGES = 0x000940CF
    FSCTL_DFS_GET_REFERRALS_EX = 0x000601B0
    FSCTL_SET_ZERO_DATA = 0x000980C8
    FSCTL_FILE_LEVEL_TRIM = 0x00098208
    FSCTL_VALIDATE_NEGOTIATE_INFO = 0x00140204


class IOCTLFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31 SMB2 IOCTL Request Flags
    A flag that indicates how to process the operation
    """

    SMB2_0_IOCTL_IS_IOCTL = 0x00000000
    SMB2_0_IOCTL_IS_FSCTL = 0x00000001


class HashVersion:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31.2 SRV_READ_HASH Request HashVersion
    The version number of the algorithm used to create the Content Information.
    """

    SRV_HASH_VER_1 = 0x00000001
    SRV_HASH_VER_2 = 0x00000002


class HashRetrievalType:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31.2 SRV_READ_HASH Request HashRetrievalType
    Indicates the nature of the Offset field in am SMB2SrvReadHashRequest
    packet.
    """

    SRV_HASH_RETRIEVE_HASH_BASED = 0x00000001
    SRV_HASH_RETRIEVE_FILE_BASED = 0x00000002


class IfCapability:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.5 NETWORK_INTERFACE_INFO Response Capability
    The capabilities of the network interface
    """

    RSS_CAPABLE = 0x00000001
    RDMA_CAPABLE = 0x00000002


class SockAddrFamily:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.5.1 SOCKADDR_STORAGE Family
    The address family of the socket.
    """

    INTER_NETWORK = 0x0002
    INTER_NETWORK_V6 = 0x0017


class SMB2IOCTLRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31 SMB2 IOCTL Request
    Send by the client to issue an implementation-specific file system control
    or device control command across the network.
    """

    COMMAND = Commands.SMB2_IOCTL

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=57)),
                ("reserved", IntField(size=2, default=0)),
                (
                    "ctl_code",
                    EnumField(
                        size=4,
                        enum_type=CtlCode,
                    ),
                ),
                ("file_id", BytesField(size=16)),
                ("input_offset", IntField(size=4, default=lambda s: self._buffer_offset_value(s))),
                (
                    "input_count",
                    IntField(
                        size=4,
                        default=lambda s: len(s["buffer"]),
                    ),
                ),
                ("max_input_response", IntField(size=4)),
                ("output_offset", IntField(size=4, default=lambda s: self._buffer_offset_value(s))),
                ("output_count", IntField(size=4, default=0)),
                ("max_output_response", IntField(size=4)),
                (
                    "flags",
                    EnumField(
                        size=4,
                        enum_type=IOCTLFlags,
                    ),
                ),
                ("reserved2", IntField(size=4, default=0)),
                ("buffer", BytesField(size=lambda s: s["input_count"].get_value())),
            ]
        )
        super().__init__()

    def _buffer_offset_value(self, structure):
        # The offset from the beginning of the SMB2 header to the value of the
        # buffer, 0 if no buffer is set
        if len(structure["buffer"]) > 0:
            header_size = 64
            request_size = structure["structure_size"].get_value()
            return header_size + request_size - 1
        else:
            return 0


class SMB2SrvCopyChunkCopy(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31.1 SRV_COPYCHUNK_COPY
    Sent in an SMB2 IOCTL Request by the client to initiate a server-side copy
    of data.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("source_key", BytesField(size=24)),
                ("chunk_count", IntField(size=4, default=lambda s: len(s["chunks"].get_value()))),
                ("reserved", IntField(size=4)),
                (
                    "chunks",
                    ListField(
                        size=lambda s: s["chunk_count"].get_value() * 24,
                        list_count=lambda s: s["chunk_count"].get_value(),
                        list_type=StructureField(size=24, structure_type=SMB2SrvCopyChunk),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2SrvCopyChunk(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31.1.1 SRC_COPYCHUNK
    Packet sent in the Chunks array of an SRC_COPY_CHUNK_COPY packet to
    describe an individual data range to copy.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("source_offset", IntField(size=8)),
                ("target_offset", IntField(size=8)),
                ("length", IntField(size=4)),
                ("reserved", IntField(size=4)),
            ]
        )
        super().__init__()


class SMB2SrvReadHashRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31.2 SRC_READ_HASH Request
    Sent by the client in an SMB2 IOCTL Request to retrieve data from the
    Content Information File associated with a specified file.
    Not valid for the SMB 2.0.2 dialect.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("hash_type", IntField(size=4, default=1)),  # SRV_HASH_TYPE_PEER_DIST
                ("hash_version", EnumField(size=4, enum_type=HashVersion)),
                ("hash_retrieval_type", EnumField(size=4, enum_type=HashRetrievalType)),
                ("length", IntField(size=4)),
                ("offset", IntField(size=8)),
            ]
        )
        super().__init__()


class SMB2SrvNetworkResiliencyRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31.3 NETWORK_RESILIENCY_REQUEST Request
    Sent by the client to request resiliency for a specified open file.
    Not valid for the SMB 2.0.2 dialect.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                # timeout is in milliseconds
                ("timeout", IntField(size=4)),
                ("reserved", IntField(size=4)),
            ]
        )
        super().__init__()


class SMB2ValidateNegotiateInfoRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.31.4 VALIDATE_NEGOTIATE_INFO Request
    Packet sent to the server to request validation of a previous SMB 2
    NEGOTIATE request.
    Only valid for the SMB 3.0 and 3.0.2 dialects.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "capabilities",
                    FlagField(
                        size=4,
                        flag_type=Capabilities,
                    ),
                ),
                ("guid", UuidField()),
                (
                    "security_mode",
                    EnumField(
                        size=2,
                        enum_type=SecurityMode,
                    ),
                ),
                ("dialect_count", IntField(size=2, default=lambda s: len(s["dialects"].get_value()))),
                (
                    "dialects",
                    ListField(
                        size=lambda s: s["dialect_count"].get_value() * 2,
                        list_count=lambda s: s["dialect_count"].get_value(),
                        list_type=EnumField(size=2, enum_type=Dialects),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2IOCTLResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32 SMB2 IOCTL Response
    Sent by the server to transmit the results of a client SMB2 IOCTL Request.
    """

    COMMAND = Commands.SMB2_IOCTL

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=49)),
                ("reserved", IntField(size=2, default=0)),
                (
                    "ctl_code",
                    EnumField(
                        size=4,
                        enum_type=CtlCode,
                    ),
                ),
                ("file_id", BytesField(size=16)),
                ("input_offset", IntField(size=4)),
                ("input_count", IntField(size=4)),
                ("output_offset", IntField(size=4)),
                ("output_count", IntField(size=4)),
                ("flags", IntField(size=4, default=0)),
                ("reserved2", IntField(size=4, default=0)),
                (
                    "buffer",
                    BytesField(
                        size=lambda s: s["output_count"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2SrvCopyChunkResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.1 SRV_COPYCHUNK_RESPONSE
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("chunks_written", IntField(size=4)),
                ("chunk_bytes_written", IntField(size=4)),
                ("total_bytes_written", IntField(size=4)),
            ]
        )
        super().__init__()


class SMB2SrvSnapshotArray(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.2 SRV_SNAPSHOT_ARRAY
    Sent by the server in response to an SMB2IOCTLResponse for the
    FSCTL_SRV_ENUMERATE_SNAPSHOTS request.
    """

    def __init__(self):
        # TODO: validate this further when working with actual snapshots
        self.fields = OrderedDict(
            [
                ("number_of_snapshots", IntField(size=4)),
                ("number_of_snapshots_returned", IntField(size=4)),
                ("snapshot_array_size", IntField(size=4)),
                ("snapshots", BytesField()),
            ]
        )
        super().__init__()


class SMB2SrvRequestResumeKey(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.3 SRV_REQUEST_RESUME_KEY Response
    Sent by the server in response to an SMB2IOCTLResponse for the
    FSCTL_SRV_REQUEST_RESUME_KEY request.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("resume_key", BytesField(size=24)),
                (
                    "context_length",
                    IntField(
                        size=4,
                        default=lambda s: len(s["context"]),
                    ),
                ),
                (
                    "context",
                    BytesField(
                        size=lambda s: s["context_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2NetworkInterfaceInfo(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.5 NETWORK_INTERFACE_INFO Response
    The NETWORK_INTERFACE_INFO returned to the client in an SMB2IOCTLResposne
    for the FSCTL_QUERY_NETWORK_INTERFACE_INFO.

    Use the pack_multiple and unpack_multiple to handle multiple interfaces
    that are returned in the SMB2IOCTLResponse
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                # 0 if no more network interfaces
                ("next", IntField(size=4)),
                ("if_index", IntField(size=4)),
                ("capability", FlagField(size=4, flag_type=IfCapability)),
                ("reserved", IntField(size=4)),
                ("link_speed", IntField(size=8)),
                ("sock_addr_storage", StructureField(size=128, structure_type=SockAddrStorage)),
            ]
        )
        super().__init__()

    @staticmethod
    def pack_multiple(messages):
        """
        Packs a list of SMB2NetworkInterfaceInfo messages and set's the next
        value accordingly. The byte value returned is then attached to the
        SMBIOCTLResponse message.

        :param messages: List of SMB2NetworkInterfaceInfo messages
        :return: bytes of the packed messages
        """
        data = b""
        msg_count = len(messages)
        for i, msg in enumerate(messages):
            if i == msg_count - 1:
                msg["next"] = 0
            else:
                msg["next"] = 152
            data += msg.pack()
        return data

    @staticmethod
    def unpack_multiple(data):
        """
        Get's a list of SMB2NetworkInterfaceInfo messages from the byte value
        passed in. This is the raw buffer value that is set on the
        SMB2IOCTLResponse message.

        :param data: bytes of the messages
        :return: List of SMB2NetworkInterfaceInfo messages
        """
        chunks = []
        while data:
            info = SMB2NetworkInterfaceInfo()
            data = info.unpack(data)
            chunks.append(info)

        return chunks


class SockAddrStorage(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.5.1 SOCKADDR_STORAGE
    Socket Address information.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("family", EnumField(size=2, enum_type=SockAddrFamily)),
                (
                    "buffer",
                    StructureField(
                        size=lambda s: self._get_buffer_size(s),
                        structure_type=lambda s: self._get_buffer_structure_type(s),
                    ),
                ),
                (
                    "reserved",
                    BytesField(
                        size=lambda s: self._get_reserved_size(s),
                        default=lambda s: b"\x00" * self._get_reserved_size(s),
                    ),
                ),
            ]
        )
        super().__init__()

    def _get_buffer_size(self, structure):
        if structure["family"].get_value() == SockAddrFamily.INTER_NETWORK:
            return 14
        else:
            return 26

    def _get_buffer_structure_type(self, structure):
        if structure["family"].get_value() == SockAddrFamily.INTER_NETWORK:
            return SockAddrIn
        else:
            return SockAddrIn6

    def _get_reserved_size(self, structure):
        if structure["family"].get_value() == SockAddrFamily.INTER_NETWORK:
            return 112
        else:
            return 100


class SockAddrIn(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.5.1.1 SOCKADDR_IN
    Socket address information for an IPv4 address
    """

    def __init__(self):
        self.fields = OrderedDict(
            [("port", IntField(size=2)), ("ipv4_address", BytesField(size=4)), ("reserved", IntField(size=8))]
        )
        super().__init__()

    def get_ipaddress(self):
        addr_bytes = self["ipv4_address"].get_value()
        return socket.inet_ntoa(addr_bytes)

    def set_ipaddress(self, address):
        # set's the ipv4 address field from the address string passed in, this
        # needs to be the full ipv4 address including periods, e.g.
        # 192.168.1.1
        addr_bytes = socket.inet_aton(address)
        self["ipv4_address"].set_value(addr_bytes)


class SockAddrIn6(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.5.1.2 SOCKADDR_IN6
    Socket address information for an IPv6 address
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("port", IntField(size=2)),
                ("flow_info", IntField(size=4)),
                ("ipv6_address", BytesField(size=16)),
                ("scope_id", IntField(size=4)),
            ]
        )
        super().__init__()

    def get_ipaddress(self):
        # get's the full IPv6 Address, note this is the full address and has
        # not been concatenated
        addr_bytes = self["ipv6_address"].get_value()
        address = binascii.hexlify(addr_bytes).decode("utf-8")
        return ":".join([address[i : i + 4] for i in range(0, len(address), 4)])

    def set_ipaddress(self, address):
        # set's the ipv6_address field from the address passed in, note this
        # needs to be the full ipv6 address,
        # e.g. fe80:0000:0000:0000:0000:0000:0000:0000 and not any short form
        address = address.replace(":", "")
        if len(address) != 32:
            raise ValueError("When setting an IPv6 address, it must be in the full form without concatenation")
        self["ipv6_address"].set_value(binascii.unhexlify(address))


class SMB2ValidateNegotiateInfoResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.32.6 VALIDATE_NEGOTIATE_INFO Response
    Packet sent by the server on a request validation of SMB 2 negotiate
    request.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "capabilities",
                    FlagField(
                        size=4,
                        flag_type=Capabilities,
                    ),
                ),
                ("guid", UuidField()),
                ("security_mode", EnumField(size=2, enum_type=SecurityMode, enum_strict=False)),
                ("dialect", EnumField(size=2, enum_type=Dialects)),
            ]
        )
        super().__init__()



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import time
from collections import OrderedDict, namedtuple
from typing import Iterator

from smbprotocol.structure import (
    BytesField,
    EnumField,
    FlagField,
    IntField,
    ListField,
    Structure,
    TextField,
)

DFSTarget = namedtuple("DFSTarget", ["target_path", "set_boundary"])
"""A TargetEntry element in the TargetList ReferralCache.

Contains the DFS target details for a referral request. This object is defined in [MS-DFSC] under
`ReferralCache TargetList`_.

Attributes:
    target_path (str): The NetworkPath value in the referral entry.
    set_boundary (bool): Whether the TARGET_SET_BOUNDARY flag was set in the referral entry.

.. _ReferralCache TargetList:
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dfsc/0f9ef683-b497-47d0-8c2c-e56e34fde2e3
"""


class DomainEntry:
    """A DomainCache entry.

    A processed domain and DC referral request.

    Args:
        referral: The DFSReferralResponse entry for a domain referral request to cache.

    Attributes:
        domain_list (List[str]): A list of known domain controller hostnames for the domain.
    """

    def __init__(self, referral: DFSReferralEntryV3):
        self.domain_list = []
        self._referral = referral
        self._start_time: float = time.time()
        self._domain_hint_idx: int | None = None

    @property
    def domain_name(self) -> str:
        """The domain DFS path."""
        return self._referral.dfs_path

    @property
    def dc_hint(self) -> str:
        """The last known good domain hostname in domain_list."""
        return self.domain_list[self._domain_hint_idx]

    @dc_hint.setter
    def dc_hint(self, value: str) -> None:
        for idx, target in enumerate(self.domain_list):
            if target == value:
                self._domain_hint_idx = idx
                break

        else:
            raise ValueError("The specific domain hint does not exist in this domain cache entry")

    @property
    def is_expired(self) -> bool:
        """Whether the hint has expired or not."""
        return ((time.time() - self._start_time) - self._referral["time_to_live"].get_value()) >= 0

    @property
    def is_valid(self) -> bool:
        """Whether the domain entry has had a DC referral response or not."""
        return self._domain_hint_idx is not None and not self.is_expired

    def process_dc_referral(self, referral: DFSReferralResponse) -> None:
        if self._domain_hint_idx is None:
            self._domain_hint_idx = 0

        for dc_entry in referral["referral_entries"].get_value():
            for dc_hostname in dc_entry.network_address:
                if dc_hostname not in self.domain_list:
                    self.domain_list.append(dc_hostname)


class ReferralEntry:
    """A ReferralCache entry.

    A parsed referral response that is cached by the client for future connection requests.

    Args:
        referral: The DFSReferralResponse to cache.
    """

    def __init__(self, referral: DFSReferralResponse):
        referrals = referral["referral_entries"].get_value()
        self._referral_header_flags = referral["referral_header_flags"]
        self._referrals: list[DFSReferralEntryV1 | DFSReferralEntryV2 | DFSReferralEntryV3] = referrals
        self._start_time: float = time.time()
        self._target_hint_idx: int = 0

    @property
    def dfs_path(self) -> str:
        return self._referrals[self._target_hint_idx].dfs_path

    @property
    def is_root(self) -> bool:
        return self._referrals[self._target_hint_idx]["server_type"].has_flag(DFSServerTypes.ROOT_TARGETS)

    @property
    def is_link(self) -> bool:
        return not self.is_root

    # @property
    # def is_interlink(self) -> bool:
    #     return False

    @property
    def is_expired(self) -> bool:
        referral = self._referrals[self._target_hint_idx]
        return ((time.time() - self._start_time) - referral["time_to_live"].get_value()) >= 0

    @property
    def target_failback(self) -> bool:
        return self._referral_header_flags.has_flag(DFSReferralHeaderFlags.TARGET_FAIL_BACK)

    @property
    def target_hint(self) -> DFSTarget:
        return self.target_list[self._target_hint_idx]

    @target_hint.setter
    def target_hint(self, value: DFSTarget) -> None:
        for idx, target in enumerate(self.target_list):
            if target == value:
                self._target_hint_idx = idx
                break
        else:
            raise ValueError("The specific target hint does not exist in this referral entry")

    @property
    def target_list(self) -> list[DFSTarget]:
        return [
            DFSTarget(
                target_path=e.network_address,
                set_boundary=e["referral_entry_flags"].has_flag(DFSReferralEntryFlags.TARGET_SET_BOUNDARY),
            )
            for e in self._referrals
        ]

    def __iter__(self) -> Iterator[DFSTarget]:
        """Iterates through the target_list with a priority being the hinted value."""
        yield self.target_list[self._target_hint_idx]

        for idx, target in enumerate(self.target_list):
            if idx == self._target_hint_idx:
                continue

            yield target


class DFSReferralRequestFlags:
    """
    [MS-DFSC] 2.2.3 REQ_GET_DFS_REFERRAL_EX

    RequestFlags
    """

    SITE_NAME = 0x00000001


class DFSReferralHeaderFlags:
    """
    [MS-DFSC] 2.2.4 RESP_GET_DFS_REFERRAL

    ReferralHeaderFlags
    """

    REFERRAL_SERVERS = 0x00000001
    STORAGE_SERVERS = 0x00000002
    TARGET_FAIL_BACK = 0x00000004


class DFSServerTypes:
    """
    [MS-DFSC] 2.2.5.1 DFS_REFERRAL_V1

    ServerType
    """

    NON_ROOT_TARGETS = 0x0000
    ROOT_TARGETS = 0x0001


class DFSReferralEntryFlags:
    """
    [MS-DFSC] 2.2.5.3 DFS_REFERRAL_V3 & 2.2.5.4 DFS_REFERRAL_V4

    ReferralEntryFlags
    """

    NAME_LIST_REFERRAL = 0x0002
    TARGET_SET_BOUNDARY = 0x0004


class DFSReferralRequest(Structure):
    """
    [MS-DFSC] 2.2.2 REQ_GET_DFS_REFERRAL

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dfsc/663c9b38-41b8-4faa-b6f6-a4576b4cea62
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("max_referral_level", IntField(size=2, default=4)),
                ("request_file_name", TextField(null_terminated=True)),
            ]
        )
        super().__init__()


class DFSReferralRequestEx(Structure):
    """
    [MS-DFSC] 2.2.3 REQ_GET_DFS_REFERRAL_EX

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dfsc/575ad372-46a7-4ed6-80e5-f87768fad9a0
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("max_referral_level", IntField(size=2, default=4)),
                ("request_flags", FlagField(size=2, flag_type=DFSReferralRequestFlags)),
                (
                    "request_data_length",
                    IntField(
                        size=4,
                        default=lambda s: 4
                        + s["request_file_name_length"].get_value()
                        + s["site_name_length"].get_value(),
                    ),
                ),
                (
                    "request_file_name_length",
                    IntField(
                        size=2,
                        default=lambda s: len(s["request_file_name"]),
                    ),
                ),
                (
                    "request_file_name",
                    TextField(
                        null_terminated=True,
                        size=lambda s: s["request_file_name_length"].get_value(),
                    ),
                ),
                (
                    "site_name_length",
                    IntField(
                        size=2,
                        default=lambda s: len(s["site_name"]),
                    ),
                ),
                (
                    "site_name",
                    TextField(
                        null_terminated=True,
                        size=lambda s: s["site_name_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class DFSReferralResponse(Structure):
    """
    [MS-DFSC] 2.2.4 RESP_GET_DFS_REFERRAL

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dfsc/bd1a7a9d-dfee-4dc6-ba37-bfeb329a7bfa
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("path_consumed", IntField(size=2)),
                ("number_of_referrals", IntField(size=2)),
                ("referral_header_flags", FlagField(size=4, flag_type=DFSReferralHeaderFlags)),
                (
                    "referral_entries",
                    ListField(
                        list_count=lambda s: s["number_of_referrals"].get_value(),
                        unpack_func=lambda s, b: self._create_dfs_referral_entry(b),
                    ),
                ),
                ("string_buffer", BytesField()),
            ]
        )
        super().__init__()

    def _create_dfs_referral_entry(self, data):
        results = []
        for _ in range(self.fields["number_of_referrals"].get_value()):
            b_version = data[:1]
            if b_version == b"\x01":
                referral_entry = DFSReferralEntryV1()

            elif b_version == b"\x02":
                referral_entry = DFSReferralEntryV2()

            else:
                referral_entry = DFSReferralEntryV3()

            data = referral_entry.unpack(data)
            results.append(referral_entry)

        # Now process the string_buffer values. Do it in reverse order as the offset value needs to be calculated based
        # on the entry size(s) that come after that entry.
        entry_offset = 0
        for referral_entry in reversed(results):
            entry_offset += referral_entry["size"].get_value()
            referral_entry.process_string_buffer(data, entry_offset)

        return results


class DFSReferralEntryV1(Structure):
    """
    [MS-DFSC] 2.2.5.1 DFS_REFERRAL_V1

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dfsc/2f7570f9-cafe-442b-ae9e-3ff119cae7d8
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("version_number", IntField(size=2, default=1)),
                ("size", IntField(size=2)),
                ("server_type", EnumField(size=2, enum_type=DFSServerTypes)),
                ("referral_entry_flags", FlagField(size=2, flag_type=DFSReferralEntryFlags)),
                ("share_name", TextField(null_terminated=True)),
            ]
        )
        super().__init__()

    @property
    def network_address(self):
        return self["share_name"].get_value()

    def process_string_buffer(self, buffer, entry_offset):
        # The V1 entry does not use a string buffer so this is a no-op.
        pass


class DFSReferralEntryV2(Structure):
    """
    [MS-DFSC] 2.2.5.2 DFS_REFERRAL_V2

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dfsc/3b6ed482-0ca4-4af1-9fd0-254d137c7af4
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("version_number", IntField(size=2, default=2)),
                ("size", IntField(size=2)),
                ("server_type", FlagField(size=2, flag_type=DFSServerTypes)),
                ("referral_entry_flags", FlagField(size=2, flag_type=DFSReferralEntryFlags)),
                ("proximity", IntField(size=4)),
                ("time_to_live", IntField(size=4)),
                ("dfs_path_offset", IntField(size=2)),
                ("dfs_alternate_path_offset", IntField(size=2)),
                ("network_address_offset", IntField(size=2)),
            ]
        )
        self.dfs_path = None
        self.dfs_alternate_path = None
        self.network_address = None
        super().__init__()

    def process_string_buffer(self, buffer, entry_offset):
        buffer_fields = ["dfs_path", "dfs_alternate_path", "network_address"]

        for field_name in buffer_fields:
            field_offset = self[f"{field_name}_offset"].get_value()
            if field_offset == 0:
                continue

            field = TextField(null_terminated=True, encoding="utf-16-le")
            field.unpack(buffer[field_offset - entry_offset :])
            setattr(self, field_name, field.get_value())


class DFSReferralEntryV3(Structure):
    """
    [MS-DFSC] 2.2.5.3 DFS_REFERRAL_V3

    This is also the same as V4 just with the addition of more referral_entry_flags and version_number being 4.

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dfsc/50a739cf-8211-4b17-9990-63696d37028d
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("version_number", IntField(size=2, default=3)),
                ("size", IntField(size=2)),
                ("server_type", FlagField(size=2, flag_type=DFSServerTypes, flag_strict=False)),
                ("referral_entry_flags", FlagField(size=2, flag_type=DFSReferralEntryFlags, flag_strict=False)),
                ("time_to_live", IntField(size=4)),
                # The follow fields depend on the ReferralEntryFlags, the actual field names are when the entry is not a
                # NameListReferral.
                ("dfs_path_offset", IntField(size=2)),  # SpecialNameOffset
                ("dfs_alternate_path_offset", IntField(size=2)),  # NumberOfExpandedNames
                ("network_address_offset", IntField(size=2)),  # ExpandedNameOffset
                ("service_site_guid", BytesField(size=lambda s: s["size"].get_value() - 18)),  # Padding
            ]
        )
        self.dfs_path = None
        self.dfs_alternate_path = None
        self.network_address = None
        super().__init__()

    def process_string_buffer(self, buffer, entry_offset):
        is_name_list = self["referral_entry_flags"].has_flag(DFSReferralEntryFlags.NAME_LIST_REFERRAL)
        buffer_fields = ["dfs_path", "network_address"]
        if not is_name_list:
            buffer_fields.insert(1, "dfs_alternate_path")

        for field_name in buffer_fields:
            field_offset = self[f"{field_name}_offset"].get_value()
            if field_offset == 0:
                continue

            string_offset = field_offset - entry_offset
            if is_name_list and field_name == "network_address":
                value = []
                for _ in range(self["dfs_alternate_path_offset"].get_value()):
                    field = TextField(null_terminated=True, encoding="utf-16-le")
                    field.unpack(buffer[string_offset:])
                    value.append(field.get_value())
                    string_offset += len(field)

            else:
                field = TextField(null_terminated=True, encoding="utf-16-le")
                field.unpack(buffer[string_offset:])
                value = field.get_value()

            setattr(self, field_name, value)



# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import hashlib
import logging
import random
from collections import OrderedDict

import spnego
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.kbkdf import KBKDFHMAC, CounterLocation, Mode

from smbprotocol import Dialects
from smbprotocol.connection import Capabilities, Ciphers, SecurityMode
from smbprotocol.exceptions import (
    MoreProcessingRequired,
    SMBAuthenticationError,
    SMBException,
)
from smbprotocol.header import Commands, NtStatus
from smbprotocol.structure import BytesField, EnumField, FlagField, IntField, Structure

log = logging.getLogger(__name__)


class SessionFlags:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.6 SMB2 SESSION_SETUP Response Flags
    Flags the indicates additional information about the session.
    """

    SMB2_SESSION_FLAG_IS_GUEST = 0x0001
    SMB2_SESSION_FLAG_IS_NULL = 0x0002
    SMB2_SESSION_FLAG_ENCRYPT_DATA = 0x0004


class SMB2SessionSetupRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.5 SMB2 SESSION_SETUP Request
    The SMB2 SESSION_SETUP Request packet is sent by the client to request a
    new authenticated session within a new or existing SMB 2 connection.
    """

    COMMAND = Commands.SMB2_SESSION_SETUP

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=25,
                    ),
                ),
                ("flags", IntField(size=1)),
                (
                    "security_mode",
                    EnumField(
                        size=1,
                        enum_type=SecurityMode,
                    ),
                ),
                (
                    "capabilities",
                    FlagField(
                        size=4,
                        flag_type=Capabilities,
                    ),
                ),
                ("channel", IntField(size=4)),
                (
                    "security_buffer_offset",
                    IntField(
                        size=2,
                        default=88,  # (header size 64) + (response size 24)
                    ),
                ),
                (
                    "security_buffer_length",
                    IntField(
                        size=2,
                        default=lambda s: len(s["buffer"]),
                    ),
                ),
                ("previous_session_id", IntField(size=8)),
                (
                    "buffer",
                    BytesField(
                        size=lambda s: s["security_buffer_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2SessionSetupResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.6 SMB2 SESSION_SETUP Response
    The SMB2 SESSION_SETUP Response packet is sent by the server in response to
    an SMB2 SESSION_SETUP Request.
    """

    COMMAND = Commands.SMB2_SESSION_SETUP

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=9,
                    ),
                ),
                (
                    "session_flags",
                    FlagField(
                        size=2,
                        flag_type=SessionFlags,
                    ),
                ),
                (
                    "security_buffer_offset",
                    IntField(
                        size=2,
                        default=72,  # (header size 64) + (response size 8)
                    ),
                ),
                (
                    "security_buffer_length",
                    IntField(
                        size=2,
                        default=lambda s: len(s["buffer"]),
                    ),
                ),
                (
                    "buffer",
                    BytesField(
                        size=lambda s: s["security_buffer_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2Logoff(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.7/8 SMB2 LOGOFF Request/Response
    Request and response to request the termination of a particular session as
    specified by the header.
    """

    COMMAND = Commands.SMB2_LOGOFF

    def __init__(self):
        self.fields = OrderedDict([("structure_size", IntField(size=2, default=4)), ("reserved", IntField(size=2))])
        super().__init__()


class Session:
    def __init__(self, connection, username=None, password=None, require_encryption=True, auth_protocol="negotiate"):
        """
        [MS-SMB2] v53.0 2017-09-15

        3.2.1.3 Per Session
        The Session object that is used to store the details for an
        authenticated SMB session. There are 4 forms of authentication that are
        supported;

        1. SSPI Auth, Windows only if pywin32 is installed. Uses either
            Kerberos or NTLM auth depending on the environment setup and can
            use the current user's credentials if none are provided here.
        2. NTLM Auth, requires the username and password
        3. Kerberos Auth, only available in certain circumstances
        4. Guest Auth, the credentials were rejected but the server allows a
            fallback to guest authentication (insecure and non-default)

        NTLM Auth is the fallback as it should be available in most scenarios
        while Kerberos only works on a system where python-gssapi is installed
        and the GGF extension for inquire_sec_context_by_oid is available
        (Linux), or pywin32 is installed (Windows).

        If using Kerberos Auth, the username and password can be omitted which
        means the default user kerb ticket (if available) is used. If the
        username is specified and not the password then it will get the kerb
        ticket for the principal specified (kinit must be used to get this
        ticket beforehand). If both the user and password are specified then
        it will get a ticket for the user instead of relying on the default
        store.

        If guest auth was negotiated based on a bad credential then signing
        and encryption is not allowed, for this to ultimately work the user
        must set require_signing=False when creating the Connection and
        require_encryption=False when creating the Session.

        :param connection: The Connection object that the session will use
        :param username: The username of the user to authenticate with
        :param password: The password of the user to authenticate with
        :param require_encryption: Whether any messages sent over the session
            require encryption regardless of the server settings (Dialects 3+),
            needs to be set to False for older dialects.
        :param auth_protocol: The protocol to use for authentication. Possible
            values are 'negotiate', 'ntlm' or 'kerberos'. Defaults to
            'negotiate'.
        """
        log.info("Initialising session with username: %s", username)
        self._connected = False
        self.session_id = 0
        self.require_encryption = require_encryption

        # Table of tree connection, lookup by TreeConnect.tree_connect_id and
        # by share_name
        self.tree_connect_table = {}

        # First 16 bytes of the cryptographic key for this authenticated
        # context, right-padded with 0 bytes
        self.session_key = None

        self.signing_required = connection.require_signing
        self.connection = connection
        self.username = username
        self.password = password

        # No need to validate this as the spnego library will raise a ValueError
        self.auth_protocol = auth_protocol

        # Table of OpenFile, lookup by OpenFile.file_id
        self.open_table = {}

        # SMB 3.x+
        # List of Channel
        self.channel_list = []

        # 16-bit identifier incremented on a network disconnect that indicates
        # to the server the client's Channel change
        self.channel_sequence = None

        self.encrypt_data = None
        self.encryption_key = None
        self.decryption_key = None
        self.signing_key = None
        self.application_key = None
        self.full_session_key = None

        # SMB 3.1.1+
        # Preauth integrity value computed for the exhange of SMB2
        # SESSION_SETUP request and response for this session
        self.preauth_integrity_hash_value = []

    def connect(self):
        log.debug("Decoding SPNEGO token containing supported auth mechanisms")
        try:
            context = spnego.client(
                self.username,
                self.password,
                service="cifs",
                hostname=self.connection.server_name,
                options=spnego.NegotiateOptions.session_key,
                protocol=self.auth_protocol,
            )
        except spnego.exceptions.SpnegoError as err:
            raise SMBAuthenticationError(f"Failed to authenticate with server: {err}") from err

        in_token = self.connection.gss_negotiate_token
        if self.auth_protocol != "negotiate":
            in_token = None  # The GSS Negotiate Token can only be used for Negotiate auth.

        while not context.complete or in_token:
            try:
                out_token = context.step(in_token)
            except spnego.exceptions.SpnegoError as err:
                raise SMBAuthenticationError(f"Failed to authenticate with server: {err}") from err

            if not out_token:
                break

            session_setup = SMB2SessionSetupRequest()
            session_setup["capabilities"] = Capabilities.SMB2_GLOBAL_CAP_DFS
            session_setup["security_mode"] = self.connection.client_security_mode
            session_setup["buffer"] = out_token

            log.info("Sending SMB2_SESSION_SETUP request message")
            request = self.connection.send(session_setup, sid=self.session_id, credit_request=64)
            request_id = request.message["message_id"].get_value()

            log.info("Receiving SMB2_SESSION_SETUP response message")
            try:
                response = self.connection.receive(request)
            except MoreProcessingRequired as exc:
                response = exc.header

            request_preauth = self.connection.preauth_integrity_session_hash_value.pop(request_id)
            self.preauth_integrity_hash_value += request_preauth

            # If this is the first time we received the actual session_id,
            # record the returned id.
            session_id = response["session_id"].get_value()
            if self.session_id == 0 and session_id:
                self.session_id = session_id

            setup_response = SMB2SessionSetupResponse()
            setup_response.unpack(response["data"].get_value())

            in_token = setup_response["buffer"].get_value()

            status = response["status"].get_value()
            if status == NtStatus.STATUS_MORE_PROCESSING_REQUIRED:
                log.info("More processing is required for SMB2_SESSION_SETUP")

        log.info("Setting session id to %s", self.session_id)
        self._connected = True

        # Move the session from the preauth table to the actual session table.
        self.connection.session_table[self.session_id] = self

        # session_key is the first 16 bytes, padded 0 if less than 16
        self.full_session_key = context.session_key
        self.session_key = self.full_session_key[:16].ljust(16, b"\x00")

        if self.connection.dialect >= Dialects.SMB_3_1_1:
            preauth_hash = b"\x00" * 64
            for hash_list in [self.connection.preauth_integrity_hash_value, self.preauth_integrity_hash_value]:
                for message in hash_list:
                    # Technically the algo is based on preauth_integrity_hash_id but we only support the 1
                    preauth_hash = hashlib.sha512(preauth_hash + message).digest()

            self.signing_key = self._smb3kdf(self.session_key, b"SMBSigningKey\x00", preauth_hash)
            self.application_key = self._smb3kdf(self.session_key, b"SMBAppKey\x00", preauth_hash)

            if self.connection.cipher_id in [
                Ciphers.AES_256_CCM,
                Ciphers.AES_256_GCM,
            ]:
                key_length = 32
                key = self.full_session_key
            else:
                key_length = 16
                key = self.session_key

            self.encryption_key = self._smb3kdf(key, b"SMBC2SCipherKey\x00", preauth_hash, length=key_length)
            self.decryption_key = self._smb3kdf(key, b"SMBS2CCipherKey\x00", preauth_hash, length=key_length)

        elif self.connection.dialect >= Dialects.SMB_3_0_0:
            self.signing_key = self._smb3kdf(self.session_key, b"SMB2AESCMAC\x00", b"SmbSign\x00")
            self.application_key = self._smb3kdf(self.session_key, b"SMB2APP\x00", b"SmbRpc\x00")
            self.encryption_key = self._smb3kdf(self.session_key, b"SMB2AESCCM\x00", b"ServerIn \x00")
            self.decryption_key = self._smb3kdf(self.session_key, b"SMB2AESCCM\x00", b"ServerOut\x00")

        else:
            self.signing_key = self.session_key
            self.application_key = self.session_key

        flags = setup_response["session_flags"]
        if flags.has_flag(SessionFlags.SMB2_SESSION_FLAG_ENCRYPT_DATA) or self.require_encryption:
            # make sure the connection actually supports encryption
            if not self.connection.supports_encryption:
                raise SMBException("SMB encryption is required but the connection does not support it")

            self.encrypt_data = True
            self.signing_required = False  # encryption covers signing

        else:
            self.encrypt_data = False

        if flags.has_flag(SessionFlags.SMB2_SESSION_FLAG_IS_GUEST) or flags.has_flag(
            SessionFlags.SMB2_SESSION_FLAG_IS_NULL
        ):
            self.session_key = None
            self.signing_key = None
            self.application_key = None
            self.encryption_key = None
            self.decryption_key = None

            if self.signing_required or self.encrypt_data:
                self.session_id = None
                raise SMBException(
                    "SMB encryption or signing was required but session was authenticated as a guest "
                    "which does not support encryption or signing"
                )

        if self.signing_required:
            log.info("Verifying the SMB Setup Session signature as auth is successful")
            self.connection.verify_signature(response, self.session_id, force=True)

    def disconnect(self, close=True):
        """
        Logs off the session

        :param close: Will close all tree connects in a session
        """
        if not self._connected:
            # already disconnected so let's return
            return

        if close:
            for open in list(self.open_table.values()):
                open.close(False)

            for tree in list(self.tree_connect_table.values()):
                tree.disconnect()

        log.info("Session: %s - Logging off of SMB Session", self.username)
        logoff = SMB2Logoff()
        log.info("Session: %s - Sending Logoff message", self.username)
        log.debug(logoff)
        request = self.connection.send(logoff, sid=self.session_id)

        log.info("Session: %s - Receiving Logoff response", self.username)
        res = self.connection.receive(request)
        res_logoff = SMB2Logoff()
        res_logoff.unpack(res["data"].get_value())
        log.debug(res_logoff)
        self._connected = False
        del self.connection.session_table[self.session_id]

    def _smb3kdf(self, ki, label, context, length=16):
        """
        See SMB 3.x key derivation function
        https://blogs.msdn.microsoft.com/openspecification/2017/05/26/smb-2-and-smb-3-security-in-windows-10-the-anatomy-of-signing-and-cryptographic-keys/

        :param ki: The session key is the KDK used as an input to the KDF
        :param label: The purpose of this derived key as bytes string
        :param context: The context information of this derived key as bytes
        :param length: The length of the key to generate
        string
        :return: Key derived by the KDF as specified by [SP800-108] 5.1
        """
        kdf = KBKDFHMAC(
            algorithm=hashes.SHA256(),
            mode=Mode.CounterMode,
            length=length,
            rlen=4,
            llen=4,
            location=CounterLocation.BeforeFixed,
            label=label,
            context=context,
            fixed=None,
            backend=default_backend(),
        )
        return kdf.derive(ki)



# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import binascii
import hashlib
import hmac
import logging
import os
import struct
import threading
import time
import typing as t
from collections import OrderedDict
from datetime import datetime
from threading import Lock

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import cmac
from cryptography.hazmat.primitives.ciphers import aead, algorithms

from smbprotocol import MAX_PAYLOAD_SIZE, Dialects
from smbprotocol._text import to_text
from smbprotocol.exceptions import (
    SMB2SymbolicLinkErrorResponse,
    SMBConnectionClosed,
    SMBException,
    SMBResponseException,
)
from smbprotocol.header import (
    Commands,
    NtStatus,
    Smb2Flags,
    SMB2HeaderAsync,
    SMB2HeaderRequest,
    SMB2HeaderResponse,
)
from smbprotocol.open import Open
from smbprotocol.structure import (
    BytesField,
    DateTimeField,
    EnumField,
    FlagField,
    IntField,
    ListField,
    Structure,
    StructureField,
    TextField,
    UuidField,
)
from smbprotocol.transport import Tcp

log = logging.getLogger(__name__)


class SecurityMode:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3 SMB2 NEGOTIATE Request SecurityMode
    Indicates whether SMB signing is enabled or required by the client.
    """

    SMB2_NEGOTIATE_SIGNING_ENABLED = 0x0001
    SMB2_NEGOTIATE_SIGNING_REQUIRED = 0x0002


class Capabilities:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3 SMB2 NEGOTIATE Request Capabilities
    Used in SMB3.x and above, used to specify the capabilities supported.
    """

    SMB2_GLOBAL_CAP_DFS = 0x00000001
    SMB2_GLOBAL_CAP_LEASING = 0x00000002
    SMB2_GLOBAL_CAP_LARGE_MTU = 0x00000004
    SMB2_GLOBAL_CAP_MULTI_CHANNEL = 0x00000008
    SMB2_GLOBAL_CAP_PERSISTENT_HANDLES = 0x00000010
    SMB2_GLOBAL_CAP_DIRECTORY_LEASING = 0x00000020
    SMB2_GLOBAL_CAP_ENCRYPTION = 0x00000040


class NegotiateContextType:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3.1 SMB2 NEGOTIATE_CONTENT Request ContextType
    Specifies the type of context in an SMB2 NEGOTIATE_CONTEXT message.
    """

    SMB2_PREAUTH_INTEGRITY_CAPABILITIES = 0x0001
    SMB2_ENCRYPTION_CAPABILITIES = 0x0002
    SMB2_COMPRESSION_CAPABILITIES = 0x0003
    SMB2_NETNAME_NEGOTIATE_CONTEXT_ID = 0x0005
    SMB2_TRANSPORT_CAPABILITIES = 0x0006
    SMB2_RDMA_TRANSFORM_CAPABILITIES = 0x0007
    SMB2_SIGNING_CAPABILITIES = 0x0008


class HashAlgorithms:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3.1.1 SMB2_PREAUTH_INTEGRITY_CAPABILITIES
    16-bit integer IDs that specify the integrity hash algorithm supported
    """

    SHA_512 = 0x0001


class Ciphers:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3.1.2 SMB2_ENCRYPTION_CAPABILITIES
    16-bit integer IDs that specify the supported encryption algorithms.
    """

    AES_128_CCM = 0x0001
    AES_128_GCM = 0x0002
    AES_256_CCM = 0x0003
    AES_256_GCM = 0x0004


class SigningAlgorithms:
    """
    [MS-SMB2] 2.2.3.1.7 SMB2_SIGNING_CAPABILITIES

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/cb9b5d66-b6be-4d18-aa66-8784a871cc10
    16-bit integer IDs that specify the supported signing algorithms.
    """

    HMAC_SHA256 = 0x0000
    AES_CMAC = 0x0001
    AES_GMAC = 0x0002


class SMB2NegotiateRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3 SMB2 Negotiate Request
    The SMB2 NEGOTIATE Request packet is used by the client to notify the
    server what dialects of the SMB2 Protocol the client understands. This is
    only used if the client explicitly sets the Dialect to use to a version
    less than 3.1.1. Dialect 3.1.1 added support for negotiate_context and
    SMB3NegotiateRequest should be used to support that.
    """

    COMMAND = Commands.SMB2_NEGOTIATE

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=36,
                    ),
                ),
                (
                    "dialect_count",
                    IntField(
                        size=2,
                        default=lambda s: len(s["dialects"].get_value()),
                    ),
                ),
                ("security_mode", FlagField(size=2, flag_type=SecurityMode)),
                ("reserved", IntField(size=2)),
                (
                    "capabilities",
                    FlagField(
                        size=4,
                        flag_type=Capabilities,
                    ),
                ),
                ("client_guid", UuidField()),
                ("client_start_time", IntField(size=8)),
                (
                    "dialects",
                    ListField(
                        size=lambda s: s["dialect_count"].get_value() * 2,
                        list_count=lambda s: s["dialect_count"].get_value(),
                        list_type=EnumField(size=2, enum_type=Dialects),
                    ),
                ),
            ]
        )

        super().__init__()


class SMB3NegotiateRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3 SMB2 Negotiate Request
    Like SMB2NegotiateRequest but with support for setting a list of
    Negotiate Context values. This is used by default and is for Dialects 3.1.1
    or greater.
    """

    COMMAND = Commands.SMB2_NEGOTIATE

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=36,
                    ),
                ),
                (
                    "dialect_count",
                    IntField(
                        size=2,
                        default=lambda s: len(s["dialects"].get_value()),
                    ),
                ),
                (
                    "security_mode",
                    FlagField(
                        size=2,
                        flag_type=SecurityMode,
                    ),
                ),
                ("reserved", IntField(size=2)),
                (
                    "capabilities",
                    FlagField(
                        size=4,
                        flag_type=Capabilities,
                    ),
                ),
                ("client_guid", UuidField()),
                (
                    "negotiate_context_offset",
                    IntField(
                        size=4,
                        default=lambda s: self._negotiate_context_offset_value(s),
                    ),
                ),
                (
                    "negotiate_context_count",
                    IntField(
                        size=2,
                        default=lambda s: len(s["negotiate_context_list"].get_value()),
                    ),
                ),
                ("reserved2", IntField(size=2)),
                (
                    "dialects",
                    ListField(
                        size=lambda s: s["dialect_count"].get_value() * 2,
                        list_count=lambda s: s["dialect_count"].get_value(),
                        list_type=EnumField(size=2, enum_type=Dialects),
                    ),
                ),
                (
                    "padding",
                    BytesField(
                        size=lambda s: self._padding_size(s),
                        default=lambda s: b"\x00" * self._padding_size(s),
                    ),
                ),
                (
                    "negotiate_context_list",
                    ListField(
                        list_count=lambda s: s["negotiate_context_count"].get_value(),
                        unpack_func=lambda s, d: self._negotiate_context_list(s, d),
                    ),
                ),
            ]
        )
        super().__init__()

    def _negotiate_context_offset_value(self, structure):
        # The offset from the beginning of the SMB2 header to the first, 8-byte
        # aligned, negotiate context
        header_size = 64
        negotiate_size = structure["structure_size"].get_value()
        dialect_size = len(structure["dialects"])
        padding_size = self._padding_size(structure)
        return header_size + negotiate_size + dialect_size + padding_size

    def _padding_size(self, structure):
        # Padding between the end of the buffer value and the first Negotiate
        # context value so that the first value is 8-byte aligned. Padding is
        # 4 is there are no dialects specified
        mod = (structure["dialect_count"].get_value() * 2) % 8
        return 0 if mod == 0 else mod

    def _negotiate_context_list(self, structure, data):
        context_count = structure["negotiate_context_count"].get_value()
        context_list = []
        for idx in range(0, context_count):
            field, data = self._parse_negotiate_context_entry(data)
            context_list.append(field)

        return context_list

    def _parse_negotiate_context_entry(self, data):
        data_length = struct.unpack("<H", data[2:4])[0]
        negotiate_context = SMB2NegotiateContextRequest()
        negotiate_context.unpack(data[: data_length + 8])
        padded_size = 8 - (data_length % 8 or 8)

        return negotiate_context, data[8 + data_length + padded_size :]


class SMB2NegotiateContextRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3.1 SMB2 NEGOTIATE_CONTEXT Request Values
    The SMB2_NEGOTIATE_CONTEXT structure is used by the SMB2 NEGOTIATE Request
    and the SMB2 NEGOTIATE Response to encode additional properties.
    """

    COMMAND = Commands.SMB2_NEGOTIATE

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "context_type",
                    EnumField(
                        size=2,
                        enum_type=NegotiateContextType,
                    ),
                ),
                (
                    "data_length",
                    IntField(
                        size=2,
                        default=lambda s: len(s["data"].get_value()),
                    ),
                ),
                ("reserved", IntField(size=4)),
                (
                    "data",
                    StructureField(
                        size=lambda s: s["data_length"].get_value(),
                        structure_type=lambda s: self._data_structure_type(s),
                    ),
                ),
                # not actually a field but each list entry must start at the 8 byte
                # alignment
                (
                    "padding",
                    BytesField(
                        size=lambda s: self._padding_size(s),
                        default=lambda s: b"\x00" * self._padding_size(s),
                    ),
                ),
            ]
        )
        super().__init__()

    def _data_structure_type(self, structure):
        con_type = structure["context_type"].get_value()
        if con_type == NegotiateContextType.SMB2_PREAUTH_INTEGRITY_CAPABILITIES:
            return SMB2PreauthIntegrityCapabilities
        elif con_type == NegotiateContextType.SMB2_ENCRYPTION_CAPABILITIES:
            return SMB2EncryptionCapabilities
        elif con_type == NegotiateContextType.SMB2_NETNAME_NEGOTIATE_CONTEXT_ID:
            return SMB2NetnameNegotiateContextId
        elif con_type == NegotiateContextType.SMB2_SIGNING_CAPABILITIES:
            return SMB2SigningCapabilities

    def _padding_size(self, structure):
        data_size = len(structure["data"])
        return 8 - (data_size % 8 or 8)


class SMB2PreauthIntegrityCapabilities(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3.1.1 SMB2_PREAUTH_INTEGRITY_CAPABILITIES
    The SMB2_PREAUTH_INTEGRITY_CAPABILITIES context is specified in an SMB2
    NEGOTIATE request by the client to indicate which preauthentication
    integrity hash algorithms it supports and to optionally supply a
    preauthentication integrity hash salt value.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "hash_algorithm_count",
                    IntField(
                        size=2,
                        default=lambda s: len(s["hash_algorithms"].get_value()),
                    ),
                ),
                (
                    "salt_length",
                    IntField(
                        size=2,
                        default=lambda s: len(s["salt"]),
                    ),
                ),
                (
                    "hash_algorithms",
                    ListField(
                        size=lambda s: s["hash_algorithm_count"].get_value() * 2,
                        list_count=lambda s: s["hash_algorithm_count"].get_value(),
                        list_type=EnumField(size=2, enum_type=HashAlgorithms),
                    ),
                ),
                (
                    "salt",
                    BytesField(
                        size=lambda s: s["salt_length"].get_value(),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2EncryptionCapabilities(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.3.1.2 SMB2_ENCRYPTION_CAPABILITIES
    The SMB2_ENCRYPTION_CAPABILITIES context is specified in an SMB2 NEGOTIATE
    request by the client to indicate which encryption algorithms the client
    supports.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "cipher_count",
                    IntField(
                        size=2,
                        default=lambda s: len(s["ciphers"].get_value()),
                    ),
                ),
                (
                    "ciphers",
                    ListField(
                        size=lambda s: s["cipher_count"].get_value() * 2,
                        list_count=lambda s: s["cipher_count"].get_value(),
                        list_type=EnumField(size=2, enum_type=Ciphers),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2NetnameNegotiateContextId(Structure):
    """
    [MS-SMB2] 2.2.3.1.4 SMB2_NETNAME_NEGOTIATE_CONTEXT_ID

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/ca6726bd-b9cf-43d9-b0bc-d127d3c993b3

    The SMB2_NETNAME_NEGOTIATE_CONTEXT_ID context is specified in an SMB2
    NEGOTIATE request to indicate the server name the client connects to.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("net_name", TextField()),
            ]
        )
        super().__init__()


class SMB2SigningCapabilities(Structure):
    """
    [MS-SMB2] 2.2.3.1.7 SMB2_SIGNING_CAPABILITIES

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/cb9b5d66-b6be-4d18-aa66-8784a871cc10

    The SMB2_SIGNING_CAPABILITIES context is specified in an SMB2 NEGOTIATE
    request by the client to indicate which signing algorithms the client supports.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "signing_algorithm_count",
                    IntField(
                        size=2,
                        default=lambda s: len(s["signing_algorithms"].get_value()),
                    ),
                ),
                (
                    "signing_algorithms",
                    ListField(
                        size=lambda s: s["signing_algorithm_count"].get_value() * 2,
                        list_count=lambda s: s["signing_algorithm_count"].get_value(),
                        list_type=EnumField(size=2, enum_type=SigningAlgorithms),
                    ),
                ),
            ]
        )
        super().__init__()


class SMB2NegotiateResponse(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.4 SMB2 NEGOTIATE Response
    The SMB2 NEGOTIATE Response packet is sent by the server to notify the
    client of the preferred common dialect.
    """

    COMMAND = Commands.SMB2_NEGOTIATE

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=65,
                    ),
                ),
                (
                    "security_mode",
                    FlagField(
                        size=2,
                        flag_type=SecurityMode,
                    ),
                ),
                (
                    "dialect_revision",
                    EnumField(
                        size=2,
                        enum_type=Dialects,
                    ),
                ),
                (
                    "negotiate_context_count",
                    IntField(
                        size=2,
                        default=lambda s: self._negotiate_context_count_value(s),
                    ),
                ),
                ("server_guid", UuidField()),
                ("capabilities", FlagField(size=4, flag_type=Capabilities)),
                ("max_transact_size", IntField(size=4)),
                ("max_read_size", IntField(size=4)),
                ("max_write_size", IntField(size=4)),
                ("system_time", DateTimeField()),
                ("server_start_time", DateTimeField()),
                (
                    "security_buffer_offset",
                    IntField(
                        size=2,
                        default=128,  # (header size 64) + (structure size 64)
                    ),
                ),
                (
                    "security_buffer_length",
                    IntField(
                        size=2,
                        default=lambda s: len(s["buffer"].get_value()),
                    ),
                ),
                (
                    "negotiate_context_offset",
                    IntField(
                        size=4,
                        default=lambda s: self._negotiate_context_offset_value(s),
                    ),
                ),
                (
                    "buffer",
                    BytesField(
                        size=lambda s: s["security_buffer_length"].get_value(),
                    ),
                ),
                (
                    "padding",
                    BytesField(
                        size=lambda s: self._padding_size(s),
                        default=lambda s: b"\x00" * self._padding_size(s),
                    ),
                ),
                (
                    "negotiate_context_list",
                    ListField(
                        list_count=lambda s: s["negotiate_context_count"].get_value(),
                        unpack_func=lambda s, d: self._negotiate_context_list(s, d),
                    ),
                ),
            ]
        )
        super().__init__()

    def _negotiate_context_count_value(self, structure):
        # If the dialect_revision is SMBv3.1.1, this field specifies the
        # number of negotiate contexts in negotiate_context_list; otherwise
        # this field must not be used and must be reserved (0).
        if structure["dialect_revision"].get_value() == Dialects.SMB_3_1_1:
            return len(structure["negotiate_context_list"].get_value())
        else:
            return None

    def _negotiate_context_offset_value(self, structure):
        # If the dialect_revision is SMBv3.1.1, this field specifies the offset
        # from the beginning of the SMB2 header to the first 8-byte
        # aligned negotiate context entry in negotiate_context_list; otherwise
        # this field must not be used and must be reserved (0).
        if structure["dialect_revision"].get_value() == Dialects.SMB_3_1_1:
            buffer_offset = structure["security_buffer_offset"].get_value()
            buffer_size = structure["security_buffer_length"].get_value()
            padding_size = self._padding_size(structure)
            return buffer_offset + buffer_size + padding_size
        else:
            return None

    def _padding_size(self, structure):
        # Padding between the end of the buffer value and the first Negotiate
        # context value so that the first value is 8-byte aligned. Padding is
        # not required if there are not negotiate contexts
        if structure["negotiate_context_count"].get_value() == 0:
            return 0

        mod = structure["security_buffer_length"].get_value() % 8
        return 0 if mod == 0 else 8 - mod

    def _negotiate_context_list(self, structure, data):
        context_count = structure["negotiate_context_count"].get_value()
        context_list = []
        for idx in range(0, context_count):
            field, data = self._parse_negotiate_context_entry(data)
            context_list.append(field)

        return context_list

    def _parse_negotiate_context_entry(self, data):
        data_length = struct.unpack("<H", data[2:4])[0]
        negotiate_context = SMB2NegotiateContextRequest()
        negotiate_context.unpack(data[: data_length + 8])
        padded_size = data_length % 8
        if padded_size != 0:
            padded_size = 8 - padded_size

        return negotiate_context, data[8 + data_length + padded_size :]


class SMB2Echo(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.28 SMB2 Echo Request/Response
    Request and response for an SMB ECHO message.
    """

    COMMAND = Commands.SMB2_ECHO

    def __init__(self):
        self.fields = OrderedDict([("structure_size", IntField(size=2, default=4)), ("reserved", IntField(size=2))])
        super().__init__()


class SMB2CancelRequest(Structure):
    """
    [MS-SMB2] 2.2.30 - SMB2 CANCEL Request
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/91913fc6-4ec9-4a83-961b-370070067e63

    The SMB2 CANCEL Request packet is sent by the client to cancel a previously sent message on the same SMB2 transport
    connection.
    """

    COMMAND = Commands.SMB2_CANCEL

    def __init__(self):
        self.fields = OrderedDict(
            [
                (
                    "structure_size",
                    IntField(
                        size=2,
                        default=4,
                    ),
                ),
                ("reserved", IntField(size=2)),
            ]
        )
        super().__init__()


class SMB2TransformHeader(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.41 SMB2 TRANSFORM_HEADER
    The SMB2 Transform Header is used by the client or server when sending
    encrypted message. This is only valid for the SMB.x dialect family.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("protocol_id", BytesField(size=4, default=b"\xfdSMB")),
                ("signature", BytesField(size=16, default=b"\x00" * 16)),
                ("nonce", BytesField(size=16)),
                ("original_message_size", IntField(size=4)),
                ("reserved", IntField(size=2, default=0)),
                ("flags", IntField(size=2, default=1)),
                ("session_id", IntField(size=8)),
                ("data", BytesField()),  # not in spec
            ]
        )
        super().__init__()


class Connection:
    def __init__(self, guid, server_name, port=445, require_signing=True):
        """
        [MS-SMB2] v53.0 2017-09-15

        3.2.1.2 Per SMB2 Transport Connection
        Used as the transport interface for a server. Some values have been
        omitted as they can be retrieved by the Server object stored in
        self.server

        :param guid: A unique guid that represents the client
        :param server_name: The server to start the connection
        :param port: The port to use for the transport, default is 445
        :param require_signing: Whether signing is required on SMB messages
            sent over this connection
        """
        log.info(
            "Initialising connection, guid: %s, require_signing: %s, server_name: %s, port: %d",
            guid,
            require_signing,
            server_name,
            port,
        )
        self.server_name = server_name
        self.port = port
        self.transport = None  # Instanciated in .connect()

        # Table of Session entries, the order is important for smbclient.
        self.session_table = OrderedDict()

        # Table of Requests that have yet to be picked up by the application,
        # it MAY contain a response from the server as well
        self.outstanding_requests = dict()

        # Table of available sequence numbers
        self.sequence_window = dict(low=0, high=1)
        self.sequence_lock = Lock()

        # Byte array containing the negotiate token and remembered for
        # authentication
        self.gss_negotiate_token = None

        self.server_guid = None
        self.max_transact_size = None
        self.max_read_size = None
        self.max_write_size = None
        self.require_signing = require_signing

        # SMB 2.1+
        self.dialect = None
        self.supports_file_leasing = None
        # just go with False as a default for Dialect 2.0.2
        self.supports_multi_credit = False
        self.client_guid = guid

        # SMB 3.x+
        self.salt = None
        self.supports_directory_leasing = None
        self.supports_multi_channel = None
        self.supports_persistent_handles = None
        self.supports_encryption = None

        # used for SMB 3.x for secure negotiate verification on tree connect
        self.negotiated_dialects = []
        self.client_capabilities = (
            Capabilities.SMB2_GLOBAL_CAP_LARGE_MTU
            | Capabilities.SMB2_GLOBAL_CAP_ENCRYPTION
            | Capabilities.SMB2_GLOBAL_CAP_DFS
        )

        self.client_security_mode = (
            SecurityMode.SMB2_NEGOTIATE_SIGNING_REQUIRED
            if require_signing
            else SecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED
        )
        self.server_security_mode = None
        self.server_capabilities = None

        # SMB 3.1.1+
        # The hashing algorithm object that was negotiated
        self.preauth_integrity_hash_id = None

        # Preauth integrity hash value computed for the SMB2 NEGOTIATE request
        # contains the messages used to compute the hash
        self.preauth_integrity_hash_value = []
        # Table of raw header bytes for unauthenticated session setup messages,
        # indexed by message_id.
        self.preauth_integrity_session_hash_value: t.Dict[int, t.List[bytes]] = {}

        # The cipher object that was negotiated
        self.cipher_id = None

        # The signing algorithm that was negotiated
        self.signing_algorithm_id = None

        # Keep track of the message processing thread's potential traceback that it may raise.
        self._t_exc = None

    def connect(self, dialect=None, timeout=60, preferred_encryption_algos=None, preferred_signing_algos=None):
        """
        Will connect to the target server and negotiate the capabilities
        with the client. Once setup, the client MUST call the disconnect()
        function to close the listener thread. This function will populate
        various connection properties that denote the capabilities of the
        server.

        If no preferred encryption or signing algorithms are specified then
        all algorithms are offered during negotiation. Older dialects may not
        be offered if a custom encryption or signing algorithm list is
        specified without the algorithm required by that dialect.

        By default the following encryption algorithms are used:

            AES_128_GCM
            AES_128_CCM (required for SMB 3.0.x)
            AES_256_GCM
            AES_256_CCM

        By default the following signing algorithms are used:

            AES_GMAC
            AES_CMAC (required for SMB 3.0.x)
            HMAC_SHA256 (required for SMB 2.x)

        :param dialect: If specified, forces the dialect that is negotiated
            with the server, if not set, then the newest dialect supported by
            the server is used up to SMB 3.1.1
        :param timeout: The timeout in seconds to wait for the initial
            negotiation process to complete
        :param preferred_encryption_algos: A list of encryption algorithm ids
            in priority order from highest to lowest. See :class:`Ciphers` for
            a list of known identifiers.
        :param preferred_signing_algos: A list of signing algorithm ids in
            priority order from highest to lowest.
            See :class:`SigningAlgorithms` for a list of known identifiers.
        """
        log.info("Setting up transport connection")
        self.transport = Tcp(self.server_name, self.port, timeout)
        self.transport.connect()
        t_worker = threading.Thread(
            target=self._process_message_thread, name=f"msg_worker-{self.server_name}:{self.port}"
        )
        t_worker.daemon = True
        t_worker.start()

        log.info("Starting negotiation with SMB server")
        enc_algos = preferred_encryption_algos or [
            Ciphers.AES_128_GCM,
            Ciphers.AES_128_CCM,
            Ciphers.AES_256_GCM,
            Ciphers.AES_256_CCM,
        ]
        sign_algos = preferred_signing_algos or [
            SigningAlgorithms.AES_GMAC,
            SigningAlgorithms.AES_CMAC,
            SigningAlgorithms.HMAC_SHA256,
        ]
        smb_response = self._send_smb2_negotiate(dialect, timeout, enc_algos, sign_algos)
        log.info("Negotiated dialect: %s", smb_response["dialect_revision"])
        self.dialect = smb_response["dialect_revision"].get_value()
        self.max_transact_size = smb_response["max_transact_size"].get_value()
        self.max_read_size = smb_response["max_read_size"].get_value()
        self.max_write_size = smb_response["max_write_size"].get_value()
        self.server_guid = smb_response["server_guid"].get_value()
        self.gss_negotiate_token = smb_response["buffer"].get_value()

        if not self.require_signing and smb_response["security_mode"].has_flag(
            SecurityMode.SMB2_NEGOTIATE_SIGNING_REQUIRED
        ):
            self.require_signing = True
        log.info("Connection require signing: %s", self.require_signing)

        capabilities = smb_response["capabilities"]
        self.server_capabilities = capabilities
        self.server_security_mode = smb_response["security_mode"].get_value()

        # SMB 2.1
        if self.dialect >= Dialects.SMB_2_1_0:
            self.supports_file_leasing = capabilities.has_flag(Capabilities.SMB2_GLOBAL_CAP_LEASING)
            self.supports_multi_credit = capabilities.has_flag(Capabilities.SMB2_GLOBAL_CAP_LARGE_MTU)

        # SMB 3.x
        if self.dialect >= Dialects.SMB_3_0_0:
            self.supports_directory_leasing = capabilities.has_flag(Capabilities.SMB2_GLOBAL_CAP_DIRECTORY_LEASING)
            self.supports_multi_channel = capabilities.has_flag(Capabilities.SMB2_GLOBAL_CAP_MULTI_CHANNEL)

            # TODO: SMB2_GLOBAL_CAP_PERSISTENT_HANDLES
            self.supports_persistent_handles = False
            self.supports_encryption = (
                capabilities.has_flag(Capabilities.SMB2_GLOBAL_CAP_ENCRYPTION) and self.dialect < Dialects.SMB_3_1_1
            )

            # TODO: Check/add server to server_list in Client Page 203

        # SMB 3.1
        if self.dialect >= Dialects.SMB_3_1_1:
            for context in smb_response["negotiate_context_list"]:
                context_type = context["context_type"].get_value()

                if context_type == NegotiateContextType.SMB2_ENCRYPTION_CAPABILITIES:
                    self.cipher_id = context["data"]["ciphers"][0]
                    self.supports_encryption = self.cipher_id != 0

                elif context_type == NegotiateContextType.SMB2_PREAUTH_INTEGRITY_CAPABILITIES:
                    self.preauth_integrity_hash_id = context["data"]["hash_algorithms"][0]

                elif context_type == NegotiateContextType.SMB2_SIGNING_CAPABILITIES:
                    self.signing_algorithm_id = context["data"]["signing_algorithms"][0]

    def disconnect(self, close=True):
        """
        Closes the connection as well as logs off any of the
        Disconnects the TCP connection and shuts down the socket listener
        running in a thread.

        :param close: Will close all sessions in the connection as well as the
            tree connections of each session.
        """
        # We cannot close the session or tree if the socket has been closed.
        if close and self.transport.connected:
            for session in list(self.session_table.values()):
                session.disconnect(True)

        log.info("Disconnecting transport connection")
        self.transport.close()

    def send(
        self, message, sid=None, tid=None, credit_request=None, message_id=None, async_id=None, force_signature=False
    ):
        """
        Will send a message to the server that is passed in. The final unencrypted header is returned to the function
        that called this.

        :param message: An SMB message structure to send.
        :param sid: A session_id that the message is sent for.
        :param tid: A tree_id object that the message is sent for.
        :param credit_request: Specifies extra credits to be requested with the SMB header.
        :param message_id: The message_id for the header, only useful for a cancel request.
        :param async_id: The async_id for the header, only useful for a cancel request.
        :param force_signature: Force signing the SMB request even if not requested by the client/server.
        :return: Request of the message that was sent.
        """
        return self._send(
            [message],
            session_id=sid,
            tree_id=tid,
            message_id=message_id,
            credit_request=credit_request,
            async_id=async_id,
            force_signature=force_signature,
        )[0]

    def send_compound(self, messages, sid, tid, related=False):
        """
        Sends multiple messages within 1 TCP request, will fail if the size of the total length exceeds the maximum of
        the transport max.

        :param messages: A list of messages to send to the server.
        :param sid: The session_id that the request is sent for.
        :param tid: A tree_id object that the message is sent for.
        :param related: Whether each message is related to each other, sets the Session, Tree, and File Id to the same
            value as the first message.
        :return: List<Request> for each request that was sent, each entry in the list is in the same order of the
            message list that was passed in.
        """
        return self._send(messages, session_id=sid, tree_id=tid, related=related)

    def receive(self, request, wait=True, timeout=None, resolve_symlinks=True):
        """
        Polls the message buffer of the TCP connection and waits until a valid
        message is received based on the message_id passed in.

        :param request: The Request object to wait get the response for
        :param wait: Wait for the final response in the case of a STATUS_PENDING response, the pending response is
            returned in the case of wait=False
        :param timeout: Set a timeout used while waiting for the final response from the server.
        :param resolve_symlinks: Set to automatically resolve symlinks in the path when opening a file or directory.
        :return: SMB2HeaderResponse of the received message
        """
        # Make sure the receiver is still active, if not this raises an exception.
        self._check_worker_running()

        start_time = time.time()
        while True:
            iter_timeout = int(max(timeout - (time.time() - start_time), 1)) if timeout is not None else None
            if not request.response_event.wait(timeout=iter_timeout):
                value = request.message["message_id"].get_value()
                raise SMBException(
                    f"Connection timeout of {timeout} seconds exceeded while waiting for a message id {value} "
                    "response from the server"
                )

            # Use a lock on the request so that in the case of a pending response we have exclusive lock on the event
            # flag and can clear it without the future pending response taking it over before we first clear the flag.
            with request.response_event_lock:
                self._check_worker_running()  # The worker may have failed while waiting for the response, check again

                response = request.response
                status = response["status"].get_value()
                if status == NtStatus.STATUS_PENDING and wait:
                    # Received a pending message, clear the response_event flag and wait again.
                    request.response_event.clear()
                    continue
                elif status == NtStatus.STATUS_STOPPED_ON_SYMLINK and resolve_symlinks:
                    # Received when we do an Open on a path that contains a symlink. Need to capture all related
                    # requests and resend the Open + others with the redirected path. First we need to resolve the
                    # symlink path. This will fail if the symlink is pointing to a location that is not in the same
                    # tree/share as the original request.

                    # First wait for the other remaining requests to be processed. Their status will also fail and we
                    # need to make sure we update the old request with the new one properly.
                    related_requests = [self.outstanding_requests[i] for i in request.related_ids]
                    [r.response_event.wait() for r in related_requests]

                    # Now create a new request with the new path the symlink points to.
                    session = self.session_table[request.session_id]
                    tree = session.tree_connect_table[request.message["tree_id"].get_value()]

                    old_create = request.get_message_data()
                    tree_share_name = tree.share_name + "\\"
                    original_path = tree_share_name + to_text(
                        old_create["buffer_path"].get_value(), encoding="utf-16-le"
                    )

                    exp = SMBResponseException(response)
                    reparse_buffer = next(e for e in exp.error_details if isinstance(e, SMB2SymbolicLinkErrorResponse))
                    new_path = reparse_buffer.resolve_path(original_path)[len(tree_share_name) :]

                    new_open = Open(tree, new_path)
                    create_req = new_open.create(
                        old_create["impersonation_level"].get_value(),
                        old_create["desired_access"].get_value(),
                        old_create["file_attributes"].get_value(),
                        old_create["share_access"].get_value(),
                        old_create["create_disposition"].get_value(),
                        old_create["create_options"].get_value(),
                        create_contexts=old_create["buffer_contexts"].get_value(),
                        send=False,
                    )[0]

                    # Now add all the related requests (if any) to send as a compound request.
                    new_msgs = [create_req] + [r.get_message_data() for r in related_requests]
                    new_requests = self.send_compound(new_msgs, session.session_id, tree.tree_connect_id, related=True)

                    # Verify that the first request was successful before updating the related requests with the new
                    # info.
                    error = None
                    try:
                        new_response = self.receive(new_requests[0], wait=wait, timeout=timeout, resolve_symlinks=True)
                    except SMBResponseException as exc:
                        # We need to make sure we fix up the remaining responses before throwing this.
                        error = exc
                    [r.response_event.wait() for r in new_requests]

                    # Update the old requests with the new response information
                    for i, old_request in enumerate([request] + related_requests):
                        del self.outstanding_requests[old_request.message["message_id"].get_value()]
                        old_request.update_request(new_requests[i])

                    if error:
                        raise error

                    return new_response
                else:
                    # now we have a retrieval request for the response, we can delete
                    # the request from the outstanding requests
                    message_id = request.message["message_id"].get_value()
                    self.outstanding_requests.pop(message_id, None)

                    if status not in [NtStatus.STATUS_SUCCESS, NtStatus.STATUS_PENDING]:
                        raise SMBResponseException(response)

                    break

        return response

    def echo(self, sid=0, timeout=60, credit_request=1):
        """
        Sends an SMB2 Echo request to the server. This can be used to request
        more credits from the server with the credit_request param.

        On a Samba server, the sid can be 0 but for a Windows SMB Server, the
        sid of an authenticated session must be passed into this function or
        else the socket will close.

        :param sid: When talking to a Windows host this must be populated with
            a valid session_id from a negotiated session
        :param timeout: The timeout in seconds to wait for the Echo Response
        :param credit_request: The number of credits to request
        :return: the credits that were granted by the server
        """
        log.info("Sending Echo request with a timeout of %d and credit request of %d" % (timeout, credit_request))

        echo_msg = SMB2Echo()
        log.debug(echo_msg)
        req = self.send(echo_msg, sid=sid, credit_request=credit_request)

        log.info("Receiving Echo response")
        response = self.receive(req, timeout=timeout)
        log.info("Credits granted from the server echo response: %d" % response["credit_response"].get_value())
        echo_resp = SMB2Echo()
        echo_resp.unpack(response["data"].get_value())
        log.debug(echo_resp)

        return response["credit_response"].get_value()

    def verify_signature(self, header, session_id, force=False):
        """
        Verifies the SMB2 Header request/response signature.

        :param header: The SMB2Header that will have its signature verified against the signing key specified.
        :param session_id: The Session Id to denote what session security verifies the message.
        :param force: Force verification of the header even if it does not match the criteria required in normal
            scenarios.
        """
        message_id = header["message_id"].get_value()
        flags = header["flags"]
        status = header["status"].get_value()
        command = header["command"].get_value()

        if not force and (
            message_id == 0xFFFFFFFFFFFFFFFF
            or not flags.has_flag(Smb2Flags.SMB2_FLAGS_SIGNED)
            or status == NtStatus.STATUS_PENDING
            or command == Commands.SMB2_SESSION_SETUP
        ):
            return

        session = self.session_table.get(session_id, None)
        if session is None:
            raise SMBException(f"Failed to find session {session_id} for message verification")

        expected = self._generate_signature(
            header.pack(),
            session.signing_key,
            message_id,
            flags.has_flag(Smb2Flags.SMB2_FLAGS_SERVER_TO_REDIR),
            command,
        )
        actual = header["signature"].get_value()
        if actual != expected:
            actual_signature = binascii.hexlify(actual).decode()
            expected_signature = binascii.hexlify(expected).decode()
            raise SMBException(
                f"Server message signature could not be verified: {actual_signature} != {expected_signature}",
            )

    def _check_worker_running(self):
        """Checks that the message worker thread is still running and raises it's exception if it has failed."""
        if self._t_exc is not None:
            self.disconnect(False)
            raise self._t_exc

        elif not self.transport.connected:
            raise SMBConnectionClosed("SMB socket was closed, cannot send or receive any more data")

    def _send(
        self,
        messages,
        session_id=None,
        tree_id=None,
        message_id=None,
        credit_request=None,
        related=False,
        async_id=None,
        force_signature=False,
    ):
        send_data = b""
        requests = []
        session = self.session_table.get(session_id, None)
        tree = None
        is_dfs = False
        if tree_id and session:
            if tree_id not in session.tree_connect_table:
                raise SMBException("Cannot find Tree with the ID %d in the session tree table" % tree_id)
            tree = session.tree_connect_table[tree_id]
            is_dfs = tree.is_dfs_share

        total_requests = len(messages)
        for i, message in enumerate(messages):
            if i == total_requests - 1:
                next_command = 0
                padding = b""
            else:
                # each compound message must start at the 8-byte boundary
                msg_length = 64 + len(message)
                mod = msg_length % 8
                padding_length = 8 - mod if mod > 0 else 0
                next_command = msg_length + padding_length
                padding = b"\x00" * padding_length

            # When running with multiple threads we need to ensure that getting the message id and adjusting the
            # sequence windows is done in a thread safe manner so we use a lock to ensure only 1 thread accesses the
            # sequence window at a time.
            with self.sequence_lock:
                sequence_window_low = self.sequence_window["low"]
                sequence_window_high = self.sequence_window["high"]
                credit_charge = self._calculate_credit_charge(message)
                credits_available = sequence_window_high - sequence_window_low
                if credit_charge > credits_available:
                    raise SMBException(
                        f"Request requires {credit_charge} credits but only {credits_available} "
                        "credits are available"
                    )

                current_id = message_id or sequence_window_low
                if message.COMMAND != Commands.SMB2_CANCEL:
                    self.sequence_window["low"] += credit_charge if credit_charge > 0 else 1

            if async_id is None:
                header = SMB2HeaderRequest()
                header["tree_id"] = tree_id or 0
            else:
                header = SMB2HeaderAsync()
                header["flags"].set_flag(Smb2Flags.SMB2_FLAGS_ASYNC_COMMAND)
                header["async_id"] = async_id

            header["credit_charge"] = credit_charge
            header["command"] = message.COMMAND
            header["credit_request"] = credit_request if credit_request else credit_charge
            header["message_id"] = current_id
            header["session_id"] = session_id
            header["data"] = message.pack()
            header["next_command"] = next_command

            if is_dfs:
                header["flags"].set_flag(Smb2Flags.SMB2_FLAGS_DFS_OPERATIONS)

            if i != 0 and related:
                header["session_id"] = b"\xff" * 8
                header["tree_id"] = b"\xff" * 4
                header["flags"].set_flag(Smb2Flags.SMB2_FLAGS_RELATED_OPERATIONS)

            if force_signature or (session and session.signing_required and session.signing_key):
                header["flags"].set_flag(Smb2Flags.SMB2_FLAGS_SIGNED)
                b_header = header.pack() + padding
                signature = self._generate_signature(b_header, session.signing_key, current_id, False, message.COMMAND)

                # To save on unpacking and re-packing, manually adjust the signature and update the request object for
                # back-referencing.
                b_header = b_header[:48] + signature + b_header[64:]
                header["signature"] = signature
            else:
                b_header = header.pack() + padding

            send_data += b_header

            if message.COMMAND == Commands.SMB2_CANCEL:
                request = self.outstanding_requests[header["message_id"].get_value()]
            else:
                request = Request(header, type(message), self, session_id=session_id)
                self.outstanding_requests[header["message_id"].get_value()] = request

            # Make sure the preauth integrity values are updated for a negotiate or session setup message.
            if message.COMMAND == Commands.SMB2_NEGOTIATE:
                self.preauth_integrity_hash_value.append(b_header)

            elif message.COMMAND == Commands.SMB2_SESSION_SETUP:
                self.preauth_integrity_session_hash_value[current_id] = [b_header]

            requests.append(request)

        if related:
            requests[0].related_ids = [r.message["message_id"].get_value() for r in requests][1:]

        if session and session.encrypt_data or tree and tree.encrypt_data:
            send_data = self._encrypt(send_data, session)

        self._check_worker_running()
        self.transport.send(send_data)
        return requests

    def _process_message_thread(self):
        try:
            while True:
                # Wait for a max of 10 minutes before sending an echo that tells the SMB server the client is still
                # available. This stops the server from closing the connection and the associated sessions on a long
                # lived connection. A brief test shows Windows kills a connection at ~16 minutes so 10 minutes is a
                # safe choice.
                # https://github.com/jborean93/smbprotocol/issues/31
                try:
                    b_msg = self.transport.recv(600)
                except TimeoutError as ex:
                    # Check if the connection has unanswered keepalive echo requests with the reserved field set.
                    # When unanswered keep alive echo exists, the server did not respond withing two times the timeout.
                    # We assume that the server connection is dead and close it.
                    for r in self.outstanding_requests.values():
                        if (
                            r.response is None
                            and r.message["command"].get_value() == Commands.SMB2_ECHO
                            and r.message["reserved"].get_value() == 1
                        ):
                            # connection will be closed in finally block
                            raise SMBConnectionClosed(
                                "Connection timed out. Server did not respond within timeout."
                            ) from ex

                    log.debug("Sending SMB2 Echo to keep connection alive")
                    for sid in self.session_table.keys():
                        req = self.send(SMB2Echo(), sid=sid)
                        # Set this reserved field to 1 as we use that internally to check whether the outstanding
                        # requests queue should be cleared in this thread or not.
                        req.message["reserved"] = 1
                    continue

                # If recv didn't return any data then the socket is considered to be closed.
                if not b_msg:
                    return

                is_encrypted = b_msg[:4] == b"\xfdSMB"
                if is_encrypted:
                    msg = SMB2TransformHeader()
                    msg.unpack(b_msg)
                    b_msg = self._decrypt(msg)

                next_command = -1
                while next_command != 0:
                    next_command = struct.unpack("<L", b_msg[20:24])[0]
                    header_length = next_command if next_command != 0 else len(b_msg)
                    b_header = b_msg[:header_length]
                    b_msg = b_msg[header_length:]

                    header = SMB2HeaderResponse()
                    header.unpack(b_header)

                    message_id = header["message_id"].get_value()
                    request = self.outstanding_requests[message_id]

                    # Typically you want to get the Session Id from the first message in a compound request but that is
                    # unreliable for async responses. Instead get the Session Id from the original request object if
                    # the Session Id is 0xFFFFFFFFFFFFFFFF.
                    # https://social.msdn.microsoft.com/Forums/en-US/a580f7bc-6746-4876-83db-6ac209b202c4/mssmb2-change-notify-response-sessionid?forum=os_fileservices
                    session_id = header["session_id"].get_value()
                    if session_id == 0xFFFFFFFFFFFFFFFF:
                        session_id = request.session_id

                    # No need to waste CPU cycles to verify the signature if we already decrypted the header.
                    if not is_encrypted:
                        self.verify_signature(header, session_id)

                    credit_response = header["credit_response"].get_value()
                    if credit_response == 0 and not self.supports_multi_credit:
                        # If the dialect does not support credits we still need to adjust our sequence window.
                        # Otherwise the credit response may be 0 in the case of compound responses and the last
                        # response contains the credits that were granted.
                        credit_response += 1

                    with self.sequence_lock:
                        self.sequence_window["high"] += credit_response

                    command = header["command"].get_value()
                    status = header["status"].get_value()
                    if command == Commands.SMB2_NEGOTIATE:
                        self.preauth_integrity_hash_value.append(b_header)

                    elif command == Commands.SMB2_SESSION_SETUP and status == NtStatus.STATUS_MORE_PROCESSING_REQUIRED:
                        self.preauth_integrity_session_hash_value[message_id].append(b_header)

                    with request.response_event_lock:
                        if header["flags"].has_flag(Smb2Flags.SMB2_FLAGS_ASYNC_COMMAND):
                            request.async_id = b_header[32:40]

                        request.response = header
                        request.response_event.set()

                        # When we send a ping in this thread we want to make sure it doesn't linger in the outstanding
                        # request queue.
                        if request.message["reserved"].get_value() == 1:
                            self.outstanding_requests.pop(message_id, None)
        except Exception as exc:
            # The exception is raised in _check_worker_running by the main thread when send/receive is called next.
            self._t_exc = exc

            # While a caller of send/receive could theoretically catch this exception, we consider any failures
            # here a fatal errors and the connection should be closed so we exit the worker thread.
            self.disconnect(False)

        finally:
            # Make sure we fire all the request events to ensure the main thread isn't waiting on a receive.
            for request in self.outstanding_requests.values():
                request.response_event.set()

    def _generate_signature(self, b_header, signing_key, message_id, response, command):
        b_header = b_header[:48] + (b"\x00" * 16) + b_header[64:]

        if self.dialect >= Dialects.SMB_3_1_1 and self.signing_algorithm_id is not None:
            sign_id = self.signing_algorithm_id

        elif self.dialect >= Dialects.SMB_3_0_0:
            sign_id = SigningAlgorithms.AES_CMAC

        else:
            sign_id = SigningAlgorithms.HMAC_SHA256

        if sign_id == SigningAlgorithms.AES_GMAC:
            message_info = 0
            if response:
                message_info |= 1

            if command == Commands.SMB2_CANCEL:
                message_info |= 2

            nonce = b"".join(
                [
                    message_id.to_bytes(8, byteorder="little"),
                    message_info.to_bytes(4, byteorder="little"),
                ]
            )
            signature = aead.AESGCM(signing_key).encrypt(nonce, b"", b_header)

        elif sign_id == SigningAlgorithms.AES_CMAC:
            c = cmac.CMAC(algorithms.AES(signing_key), backend=default_backend())
            c.update(b_header)
            signature = c.finalize()

        else:
            hmac_algo = hmac.new(signing_key, msg=b_header, digestmod=hashlib.sha256)
            signature = hmac_algo.digest()[:16]

        return signature

    def _encrypt(self, b_data, session):
        header = SMB2TransformHeader()
        header["original_message_size"] = len(b_data)
        header["session_id"] = session.session_id

        encryption_key = session.encryption_key
        if self.dialect >= Dialects.SMB_3_1_1:
            cipher_id = self.cipher_id
        else:
            cipher_id = Ciphers.AES_128_CCM

        if cipher_id in [Ciphers.AES_128_GCM, Ciphers.AES_256_GCM]:
            cipher = aead.AESGCM
            nonce = os.urandom(12)
            header["nonce"] = nonce + (b"\x00" * 4)
        else:
            cipher = aead.AESCCM
            nonce = os.urandom(11)
            header["nonce"] = nonce + (b"\x00" * 5)

        cipher_text = cipher(encryption_key).encrypt(nonce, b_data, header.pack()[20:])
        signature = cipher_text[-16:]
        enc_message = cipher_text[:-16]

        header["signature"] = signature
        header["data"] = enc_message

        return header

    def _decrypt(self, message):
        value = message["flags"].get_value()
        if value != 0x0001:
            error_msg = f"Expecting flag of 0x0001 but got {value:x} in the SMB Transform Header Response"
            raise SMBException(error_msg)

        session_id = message["session_id"].get_value()
        session = self.session_table.get(session_id, None)
        if session is None:
            raise SMBException(f"Failed to find valid session {session_id} for message decryption")

        if self.dialect >= Dialects.SMB_3_1_1:
            cipher_id = self.cipher_id
        else:
            cipher_id = Ciphers.AES_128_CCM

        if cipher_id in [Ciphers.AES_128_GCM, Ciphers.AES_256_GCM]:
            cipher = aead.AESGCM
            nonce_length = 12
        else:
            cipher = aead.AESCCM
            nonce_length = 11

        nonce = message["nonce"].get_value()[:nonce_length]
        signature = message["signature"].get_value()
        enc_message = message["data"].get_value() + signature

        c = cipher(session.decryption_key)
        dec_message = c.decrypt(nonce, enc_message, message.pack()[20:52])
        return dec_message

    def _send_smb2_negotiate(self, dialect, timeout, encryption_algorithms, signing_algorithms):
        self.salt = os.urandom(32)

        if dialect is None:
            neg_req = SMB3NegotiateRequest()
            negotiated_dialects = [
                Dialects.SMB_2_0_2,
                Dialects.SMB_2_1_0,
                Dialects.SMB_3_0_0,
                Dialects.SMB_3_0_2,
                Dialects.SMB_3_1_1,
            ]

            if SigningAlgorithms.HMAC_SHA256 not in signing_algorithms:
                if Dialects.SMB_2_0_2 in negotiated_dialects:
                    negotiated_dialects.remove(Dialects.SMB_2_0_2)
                if Dialects.SMB_2_1_0 in negotiated_dialects:
                    negotiated_dialects.remove(Dialects.SMB_2_1_0)

            if SigningAlgorithms.AES_CMAC not in signing_algorithms or Ciphers.AES_128_CCM not in encryption_algorithms:
                if Dialects.SMB_3_0_0 in negotiated_dialects:
                    negotiated_dialects.remove(Dialects.SMB_3_0_0)
                if Dialects.SMB_3_0_2 in negotiated_dialects:
                    negotiated_dialects.remove(Dialects.SMB_3_0_2)
        else:
            if dialect >= Dialects.SMB_3_1_1:
                neg_req = SMB3NegotiateRequest()
            else:
                neg_req = SMB2NegotiateRequest()
            negotiated_dialects = [dialect]

        highest_dialect = sorted(negotiated_dialects)[-1]
        self.negotiated_dialects = neg_req["dialects"] = negotiated_dialects
        log.info(
            "Negotiating with SMB2 protocol with highest client dialect of: %s",
            [dialect for dialect, v in vars(Dialects).items() if v == highest_dialect][0],
        )

        neg_req["security_mode"] = self.client_security_mode

        if highest_dialect >= Dialects.SMB_2_1_0:
            log.debug("Adding client guid %s to negotiate request", self.client_guid)
            neg_req["client_guid"] = self.client_guid

        else:
            # Must be None, this value is used to verify the negotiation info.
            self.client_guid = None

        if highest_dialect >= Dialects.SMB_3_0_0:
            log.debug("Adding client capabilities %d to negotiate request", self.client_capabilities)
            neg_req["capabilities"] = self.client_capabilities

        else:
            # Must be 0, this value is used to verify the negotiation info.
            self.client_capabilities = 0

        if highest_dialect >= Dialects.SMB_3_1_1:
            int_cap = SMB2NegotiateContextRequest()
            int_cap["context_type"] = NegotiateContextType.SMB2_PREAUTH_INTEGRITY_CAPABILITIES
            int_cap["data"] = SMB2PreauthIntegrityCapabilities()
            int_cap["data"]["hash_algorithms"] = [HashAlgorithms.SHA_512]
            int_cap["data"]["salt"] = self.salt
            log.debug(
                "Adding preauth integrity capabilities of hash SHA512 and salt %s to negotiate request",
                self.salt,
            )

            enc_cap = SMB2NegotiateContextRequest()
            enc_cap["context_type"] = NegotiateContextType.SMB2_ENCRYPTION_CAPABILITIES
            enc_cap["data"] = SMB2EncryptionCapabilities()
            supported_ciphers = encryption_algorithms
            enc_cap["data"]["ciphers"] = supported_ciphers
            log.debug("Adding encryption capabilities of AES128|256 GCM and AES128|256 CCM to negotiate request")

            netname_id = SMB2NegotiateContextRequest()
            netname_id["context_type"] = NegotiateContextType.SMB2_NETNAME_NEGOTIATE_CONTEXT_ID
            netname_id["data"] = SMB2NetnameNegotiateContextId()
            netname_id["data"]["net_name"] = self.server_name
            log.debug(f"Adding netname context id of {self.server_name} to negotiate request")

            signing_cap = SMB2NegotiateContextRequest()
            signing_cap["context_type"] = NegotiateContextType.SMB2_SIGNING_CAPABILITIES
            signing_cap["data"] = SMB2SigningCapabilities()
            signing_cap["data"]["signing_algorithms"] = signing_algorithms
            log.debug("Adding signing algorithms AES_GMAC, AES_CMAC, and HMAC_SHA256 to negotiate request")

            # remove extra padding for last list entry
            signing_cap["padding"].size = 0
            signing_cap["padding"] = b""

            neg_req["negotiate_context_list"] = [
                int_cap,
                enc_cap,
                netname_id,
                signing_cap,
            ]

        log.info("Sending SMB2 Negotiate message")
        log.debug(neg_req)
        request = self.send(neg_req)

        response = self.receive(request, timeout=timeout)
        log.info("Receiving SMB2 Negotiate response")
        log.debug(response)

        smb_response = SMB2NegotiateResponse()
        smb_response.unpack(response["data"].get_value())

        return smb_response

    def _calculate_credit_charge(self, message):
        """
        Calculates the credit charge for a request based on the command. If
        connection.supports_multi_credit is not True then the credit charge
        isn't valid so it returns 0.

        The credit charge is the number of credits that are required for
        sending/receiving data over 64 kilobytes, in the existing messages only
        the Read, Write, Query Directory or IOCTL commands will end in this
        scenario and each require their own calculation to get the proper
        value. The generic formula for calculating the credit charge is

        https://msdn.microsoft.com/en-us/library/dn529312.aspx
        (max(SendPayloadSize, Expected ResponsePayloadSize) - 1) / 65536 + 1

        :param message: The message being sent
        :return: The credit charge to set on the header
        """
        if (not self.supports_multi_credit) or (message.COMMAND == Commands.SMB2_CANCEL):
            return 0

        elif message.COMMAND == Commands.SMB2_READ:
            payload_size = message["length"].get_value() + message["read_channel_info_length"].get_value()

        elif message.COMMAND == Commands.SMB2_WRITE:
            payload_size = message["length"].get_value() + message["write_channel_info_length"].get_value()

        elif message.COMMAND == Commands.SMB2_IOCTL:
            max_in_size = len(message["buffer"])
            max_out_size = message["max_output_response"].get_value()
            payload_size = max(max_in_size, max_out_size)

        elif message.COMMAND == Commands.SMB2_QUERY_DIRECTORY:
            max_in_size = len(message["buffer"])
            max_out_size = message["output_buffer_length"].get_value()
            payload_size = max(max_in_size, max_out_size)

        else:
            payload_size = 1

        credit_charge = (max(0, payload_size - 1) // MAX_PAYLOAD_SIZE) + 1
        return credit_charge


class Request:
    def __init__(self, message, message_type, connection, session_id=None):
        """
        [MS-SMB2] v53.0 2017-09-15

        3.2.1.7 Per Pending Request
        For each request that was sent to the server and is await a response
        :param message: The message to be sent in the request
        :param message_type: The type of message that is set in the header's data field.
        :param connection: The Connection the request was sent under.
        :param session_id: The Session Id the request was for.
        """
        self.async_id = None
        self.message = message
        self.timestamp = datetime.now()
        self.cancelled = False

        # Used to contain the corresponding response from the server as the receiving in done in a separate thread.
        self.response = None

        # Used by the recv processing thread to say the response has been received and is ready for consumption.
        self.response_event = threading.Event()

        # Used to lock the request when the main thread is processing the PENDING result in case the background thread
        # receives the final result and fires the event before main clears it.
        self.response_event_lock = threading.Lock()

        # Stores the message_ids of related messages that are sent in a compound request. This is only set on the 1st
        # message in the request. Used when STATUS_STOPPED_ON_SYMLINK is set and we need to send the whole compound
        # request again with the new path.
        self.related_ids = []

        # Cannot rely on the message values as it could be a related compound msg which does not set these values.
        self.session_id = session_id

        self._connection = connection
        self._message_type = message_type  # Used to rehydrate the message data in case it's needed again.

    def cancel(self):
        if self.cancelled is True:
            return

        message_id = self.message["message_id"].get_value()
        log.info("Cancelling message %s", message_id)
        self._connection.send(
            SMB2CancelRequest(), sid=self.session_id, credit_request=0, message_id=message_id, async_id=self.async_id
        )
        self.cancelled = True

    def get_message_data(self):
        message_obj = self._message_type()
        message_obj.unpack(self.message["data"].get_value())
        return message_obj

    def update_request(self, new_request):
        self.async_id = new_request.async_id
        self.message = new_request.message
        self.timestamp = new_request.timestamp
        self.response = new_request.response
        self.response_event = new_request.response_event
        self.response_event_lock = new_request.response_event_lock
        self.related_ids = new_request.related_ids



# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import errno
import stat as py_stat

from smbclient._os import stat
from smbprotocol.exceptions import SMBLinkRedirectionError, SMBOSError


def exists(path, **kwargs):
    """
    Return True if path refers to an existing path. Returns False for broken symbolic links or links pointing to
    unreachable locations.

    :param path: The path to check if it exists.
    :param kwargs: Common arguments used to build the SMB Session.
    :return: Bool as to whether the path exists or not.
    """
    return _exists(path, False, True, **kwargs)


def lexists(path, **kwargs):
    """
    Return True if path refers to an existing path. Returns True for broken symbolic links or links pointing to
    unreachable locations.

    :param path: The path to check if it exists.
    :param kwargs: Common arguments used to build the SMB Session.
    :return: Bool as to whether the path exists or not.
    """
    return _exists(path, True, False, **kwargs)


def getatime(path, **kwargs):
    """
    Return the time of last access of path. The return value is a floating point number giving the number of seconds
    since the epoch (see the time module). Raise OSError if the file does not exist or is inaccessible.

    :param path: The path to get the atime for.
    :param kwargs: Common arguments used to build the SMB Session.
    :return: float that represents the number of seconds since epoch for atime.
    """
    return stat(path, **kwargs).st_atime


def getmtime(path, **kwargs):
    """
    Return the time of last modification of path. The return value is a floating point number giving the number of
    seconds since the epoch (see the time module). Raise OSError if the file does not exist or is inaccessible.

    :param path: The path to get the mtime for.
    :param kwargs: Common arguments used to build the SMB Session.
    :return: float that represents the number of seconds since epoch for atime.
    """
    return stat(path, **kwargs).st_mtime


def getctime(path, **kwargs):
    """
    Return the system’s ctime which is the creation time for path. The return value is a number giving the number of
    seconds since the epoch (see the time module). Raise OSError if the file does not exist or is inaccessible.

    :param path: The path to get the ctime for.
    :param kwargs: Common arguments used to build the SMB Session.
    :return: float that represents the number of seconds since epoch for atime.
    """
    return stat(path, **kwargs).st_ctime


def getsize(path, **kwargs):
    """
    Return the size, in bytes, of path. Raise OSError if the file does not exist or is inaccessible.

    :param path: The path to get the size for.
    :param kwargs: Common arguments used to build the SMB Session.
    :return: The byte size of the path.
    """
    return stat(path, **kwargs).st_size


def isfile(path, **kwargs):
    """
    Return True if path is an existing regular file. This follows symbolic links, so both islink() and isfile() can be
    true for the same path.

    :param path: The path to check.
    :param kwargs: Common arguments used to build the SMB Session.
    :return: True if the path is a file or points to a file.
    """
    return _stat_ismode(path, py_stat.S_ISREG, True, **kwargs)


def isdir(path, **kwargs):
    """
    Return True if path is an existing directory. This follows symbolic links, so both islink() and isdir() can be true
    for the same path.

    :param path: The path to check.
    :param kwargs: Common arguments used to build the SMB Session.
    :return: True if path is a dir or points to a dir.
    """
    return _stat_ismode(path, py_stat.S_ISDIR, True, **kwargs)


def islink(path, **kwargs):
    """
    Return True if path is a symbolic link.

    :param path: The path to check
    :param kwargs: Common arguments used to build the SMB Session.
    :return: True if path is a symlink.
    """
    return _stat_ismode(path, py_stat.S_ISLNK, False, **kwargs)


def samefile(path1, path2, **kwargs):
    """
    Return True if both pathname arguments refer to the same file or directory. This is determined by the device number
    and i-node number and raises an exception if an os.stat() call on either pathname fails.

    :param path1: The first path to compare.
    :param path2: The second path to compare.
    :param kwargs: Common arguments used to build the SMB Session.
    :return: Bool that indicates whether the 2 files are the same.
    """
    stat1 = stat(path1, **kwargs)
    stat2 = stat(path2, **kwargs)
    return stat1.st_ino == stat2.st_ino and stat1.st_dev == stat2.st_dev


def _exists(path, symlink_default, follow_symlinks, **kwargs):
    try:
        stat(path, follow_symlinks=follow_symlinks, **kwargs)
        return True
    except OSError as err:
        if err.errno == errno.ENOENT:
            return False
        raise
    except SMBLinkRedirectionError:
        # Link points to another server or local drive, return false
        return symlink_default


def _stat_ismode(path, check, follow, **kwargs):
    try:
        return check(stat(path, follow_symlinks=follow, **kwargs).st_mode)
    except SMBOSError as err:
        if err.errno == errno.ENOENT:
            return False
        raise



# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import io
import logging

from smbclient._pool import ClientConfig, dfs_request, get_smb_tree
from smbprotocol import MAX_PAYLOAD_SIZE
from smbprotocol.exceptions import (
    NoMoreFiles,
    ObjectNameNotFound,
    ObjectPathNotFound,
    PathNotCovered,
    PipeBroken,
    SMBOSError,
    SMBResponseException,
)
from smbprotocol.file_info import FileAttributes, FileEndOfFileInformation
from smbprotocol.ioctl import IOCTLFlags, SMB2IOCTLRequest, SMB2IOCTLResponse
from smbprotocol.open import (
    CreateDisposition,
    CreateOptions,
    FilePipePrinterAccessMask,
    ImpersonationLevel,
    Open,
    QueryDirectoryFlags,
    ShareAccess,
    SMB2CreateRequest,
    SMB2QueryInfoRequest,
    SMB2QueryInfoResponse,
    SMB2SetInfoRequest,
    SMB2SetInfoResponse,
)

log = logging.getLogger(__name__)


SEEK_SET = io.SEEK_SET
SEEK_CUR = io.SEEK_CUR
SEEK_END = io.SEEK_END


def _parse_share_access(raw, mode):
    share_access = 0
    if raw:
        share_access_map = {
            "r": ShareAccess.FILE_SHARE_READ,
            "w": ShareAccess.FILE_SHARE_WRITE,
            "d": ShareAccess.FILE_SHARE_DELETE,
        }
        for c in raw:
            access_val = share_access_map.get(c)
            if access_val is None:
                chars = ", ".join(sorted(share_access_map.keys()))
                raise ValueError(f"Invalid share_access char {c}, can only be {chars}")
            share_access |= access_val

        if "r" in mode:
            share_access |= ShareAccess.FILE_SHARE_READ

    return share_access


def _parse_mode(raw, invalid=""):
    create_disposition = 0
    disposition_map = {
        "r": CreateDisposition.FILE_OPEN,
        "w": CreateDisposition.FILE_OVERWRITE_IF,
        "x": CreateDisposition.FILE_CREATE,
        "a": CreateDisposition.FILE_OPEN_IF,
        # These have no bearing on the CreateDisposition but are here so ValueError isn't raised
        "+": 0,
        "b": 0,
        "t": 0,
    }
    if invalid:
        for invalid_char in invalid:
            del disposition_map[invalid_char]

    for c in raw:
        dispo_val = disposition_map.get(c)
        if dispo_val is None:
            chars = ", ".join(sorted(disposition_map.keys()))
            raise ValueError(
                f"Invalid mode char {c}, can only be {chars}",
            )

        create_disposition |= dispo_val

    if create_disposition == 0:
        raise ValueError(f"Invalid mode value {raw}, must contain at least r, w, x, or a")

    return create_disposition


def _chunk_size(connection, length, operation):
    """
    Get the maximum size of data we can read/write. Also gets the number of credits to request to optimize subsequent
    read/write operations for the remaining length.

    :param connection: The SMB connection.
    :param length: The length of the data we are working with.
    :param operation: The operation the chunk is for: 'read', 'write', 'transact'
    :return: The size of the chunk we can use and the number of credits to request for the next operation.
    """
    max_size = getattr(connection, f"max_{operation}_size")

    # Determine the maximum data length we can send for the operation. We do this by checking the available credits and
    # calculating whatever is the smallest; length, negotiated operation size, available credit size).
    available_credits = connection.sequence_window["high"] - connection.sequence_window["low"]
    chunk_size = min(length, max_size, available_credits * MAX_PAYLOAD_SIZE)

    # Determine how many credits we need to fully optimize subsequent calls for the remaining amount of data. Basically
    # how many credits we need to send either the remaining data or the max operation size.
    remaining_length = min(max(0, length - chunk_size), max_size)
    consumed_credits = (max(0, chunk_size - 1) // MAX_PAYLOAD_SIZE) + 1
    remaining_credits = (max(0, remaining_length - 1) // MAX_PAYLOAD_SIZE) + 1
    desired_credits = max(0, remaining_credits - available_credits - consumed_credits)

    credit_request = consumed_credits + desired_credits

    return chunk_size, int(credit_request)


def _resolve_dfs(raw_io):
    """
    Resolves a DFS path for a failed Open request.

    :param raw_io: The SMBRawIO to resolve the DFS path for.
    :return: A new Open for each DFS target that was resolved.
    """
    if not raw_io.fd.tree_connect.is_dfs_share:
        return

    # Path is on a DFS root that is linked to another server.
    client_config = ClientConfig()
    raw_path = raw_io.name

    referral = dfs_request(raw_io.fd.tree_connect, raw_path[1:])
    client_config.cache_referral(referral)
    info = client_config.lookup_referral([p for p in raw_path.split("\\") if p])
    if not info:
        raise ObjectPathNotFound()

    connection_kwargs = getattr(raw_io, f"_{SMBRawIO.__name__}__kwargs", {})

    for target in info:
        new_path = raw_path.replace(info.dfs_path, target.target_path, 1)

        try:
            tree, fd_path = get_smb_tree(new_path, **connection_kwargs)

        except SMBResponseException as link_exc:
            log.warning("Failed to connect to DFS link target %s", target, exc_info=link_exc)
            continue

        # Record the target that worked for future reference.
        info.target_hint = target
        yield Open(tree, fd_path)


def ioctl_request(transaction, ctl_code, output_size=0, flags=IOCTLFlags.SMB2_0_IOCTL_IS_IOCTL, input_buffer=b""):
    """
    Sends an IOCTL request to the server.

    :param transaction: The SMBFileTransaction the request is to run under.
    :param ctl_code: The IOCTL Code of the request.
    :param output_size: Specify the max output response allowed from the server.
    :param flags: Specify custom flags to be set on the IOCTL request.
    :param input_buffer: Specify an optional input buffer for the request.
    """
    ioctl_req = SMB2IOCTLRequest()
    ioctl_req["ctl_code"] = ctl_code
    ioctl_req["file_id"] = transaction.raw.fd.file_id
    ioctl_req["max_output_response"] = output_size
    ioctl_req["flags"] = flags
    ioctl_req["buffer"] = input_buffer

    def _receive_resp(request):
        response = transaction.raw.fd.connection.receive(request)
        query_resp = SMB2IOCTLResponse()
        query_resp.unpack(response["data"].get_value())
        return query_resp["buffer"].get_value()

    transaction += (ioctl_req, _receive_resp)


def query_info(transaction, info_class, flags=0, output_buffer_length=None):
    """
    Sends a QUERY INFO request to the server for the information class passed in.

    :param transaction: The SMBFileTransaction the request is to run under.
    :param info_class: The required information class type defined in file_info.py that is being requested.
    :param flags: Optional flags to set on the query request.
    :param output_buffer_length: Override the max output buffer length, defaults to the size of the information class.
    """
    info_obj = info_class()
    query_req = SMB2QueryInfoRequest()
    query_req["info_type"] = info_obj.INFO_TYPE
    query_req["file_info_class"] = info_obj.INFO_CLASS
    query_req["file_id"] = transaction.raw.fd.file_id
    query_req["output_buffer_length"] = len(info_obj) if output_buffer_length is None else output_buffer_length
    query_req["flags"] = flags

    def _receive_resp(request):
        response = transaction.raw.fd.connection.receive(request)
        query_resp = SMB2QueryInfoResponse()
        query_resp.unpack(response["data"].get_value())
        return query_resp.parse_buffer(info_class)

    transaction += (query_req, _receive_resp)


def set_info(transaction, info_buffer):
    """
    Sends a SET INFO request to the server with the information class input.

    :param transaction: The SMBFileTransaction the request is to run under.
    :param info_buffer: The input information class to set.
    """
    set_req = SMB2SetInfoRequest()
    set_req["info_type"] = info_buffer.INFO_TYPE
    set_req["file_info_class"] = info_buffer.INFO_CLASS
    set_req["file_id"] = transaction.raw.fd.file_id
    set_req["buffer"] = info_buffer

    def _receive_resp(request):
        response = transaction.raw.fd.connection.receive(request)
        set_resp = SMB2SetInfoResponse()
        set_resp.unpack(response["data"].get_value())
        return set_resp

    transaction += (set_req, _receive_resp)


class SMBFileTransaction:
    def __init__(self, raw):
        """
        Stores compound requests in 1 class that can be committed when required. Either uses the opened raw object or
        if that is not opened, opens and closes it in the 1 request.

        :param raw: The SMBRawIO object to run the compound request with.
        """
        self.raw = raw
        self.results = None
        self._actions = []
        self._attempted_dfs_paths = set()

    def __add__(self, other):
        send_msg = other[0]
        unpack_func = other[1]

        self._actions.append((send_msg, unpack_func))
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.commit()

    def commit(self):
        """
        Sends a compound request to the server. Optionally opens and closes a handle in the same request if the handle
        is not already opened.
        """
        remove_index = []
        if self.raw.closed and (not self._actions or not isinstance(self._actions[0][0], SMB2CreateRequest)):
            self.raw.open(transaction=self)
            self._actions.insert(0, self._actions.pop(-1))  # Need to move the create to the start
            remove_index.insert(0, 0)

            self.raw.close(transaction=self)
            remove_index.insert(0, len(self._actions) - 1)

        send_msgs = []
        unpack_functions = []
        for action in self._actions:
            send_msgs.append(action[0])
            unpack_functions.append(action[1])

        sid = self.raw.fd.tree_connect.session.session_id
        tid = self.raw.fd.tree_connect.tree_connect_id
        requests = self.raw.fd.connection.send_compound(send_msgs, sid, tid, related=True)

        # Due to a wonderful threading issue we need to ensure we call .receive() for each message we sent so cannot
        # just enumerate the list in 1 line in case it throws an exception.
        failures = []
        responses = []
        try_again = False
        for idx, func in enumerate(unpack_functions):
            res = None

            try:
                try:
                    res = func(requests[idx])

                except (PathNotCovered, ObjectNameNotFound, ObjectPathNotFound):
                    # The MS-DFSC docs state that STATUS_PATH_NOT_COVERED is used when encountering a link to a
                    # different server but Samba seems to return the generic name or path not found.

                    # If the first action is not a CreateRequest then we can't resolve the DFS path in the transaction.
                    if not (idx == 0 and isinstance(send_msgs[0], SMB2CreateRequest)):
                        raise

                    for smb_open in _resolve_dfs(self.raw):
                        if smb_open.tree_connect.share_name == self.raw.fd.tree_connect.share_name:
                            continue

                        # Ensure we don't continuously try the same DFS referral targets if it's already been attempted.
                        # https://github.com/jborean93/smbprotocol/issues/228
                        tested_path = f"{smb_open.tree_connect.share_name}{smb_open.file_name}".lower()
                        if tested_path in self._attempted_dfs_paths:
                            continue

                        self._attempted_dfs_paths.add(tested_path)

                        self.raw.fd = smb_open

                        # In case this is a transaction with an explicit open we want to reopen it with the new params
                        # before trying it again.
                        self.raw.open(transaction=self)
                        self._actions[0] = self._actions.pop(-1)

                        try_again = True
                        break

                    else:
                        # Either there wasn't any DFS referrals or none of them worked, just reraise the error.
                        raise

            except SMBResponseException as exc:
                failures.append(SMBOSError(exc.status, self.raw.name))

            finally:
                responses.append(res)

        # Remove the existing open/close actions this transaction added internally.
        for idx in remove_index:
            del self._actions[idx]
            del responses[idx]

        if try_again:
            # If we updated the SMB Tree due to a DFS referral hit then try again.
            self.commit()

        elif failures:
            # If there was a failure, just raise the first exception found.
            raise failures[0]

        else:
            self.results = tuple(responses)


class SMBRawIO(io.RawIOBase):
    FILE_TYPE = None  # 'file', 'dir', or None (for unknown)
    _INVALID_MODE = ""

    def __init__(
        self, path, mode="r", share_access=None, desired_access=None, file_attributes=None, create_options=0, **kwargs
    ):
        tree, fd_path = get_smb_tree(path, **kwargs)

        self.share_access = share_access
        self.fd = Open(tree, fd_path)
        self._mode = mode
        self._name = path
        self._offset = 0
        self._flush = False
        self.__kwargs = kwargs  # Used in open for DFS referrals

        if desired_access is None:
            desired_access = 0

            # While we can open a directory, the values for FilePipePrinterAccessMask also apply to Dirs so just use
            # the same enum to simplify code.
            if "r" in self.mode or "+" in self.mode:
                desired_access |= (
                    FilePipePrinterAccessMask.FILE_READ_DATA
                    | FilePipePrinterAccessMask.FILE_READ_ATTRIBUTES
                    | FilePipePrinterAccessMask.FILE_READ_EA
                )
            if "w" in self.mode or "x" in self.mode or "a" in self.mode or "+" in self.mode:
                desired_access |= (
                    FilePipePrinterAccessMask.FILE_WRITE_DATA
                    | FilePipePrinterAccessMask.FILE_WRITE_ATTRIBUTES
                    | FilePipePrinterAccessMask.FILE_WRITE_EA
                )
        self._desired_access = desired_access

        if file_attributes is None:
            file_attributes = (
                FileAttributes.FILE_ATTRIBUTE_DIRECTORY
                if self.FILE_TYPE == "dir"
                else FileAttributes.FILE_ATTRIBUTE_NORMAL
            )
        self._file_attributes = file_attributes

        self._create_options = create_options
        self._create_options |= {
            "dir": CreateOptions.FILE_DIRECTORY_FILE,
            "file": CreateOptions.FILE_NON_DIRECTORY_FILE,
        }.get(self.FILE_TYPE, 0)

        super().__init__()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def closed(self):
        return not self.fd.connected

    @property
    def mode(self):
        return self._mode

    @property
    def name(self):
        return self._name

    def close(self, transaction=None):
        if transaction:
            transaction += self.fd.close(send=False)
        else:
            self.fd.close()

    def flush(self):
        if self._flush and self.FILE_TYPE != "pipe":
            self.fd.flush()

    def open(self, transaction=None):
        if not self.closed:
            return

        open_at_end = False
        if not transaction:
            # The SMBFileTransaction has special logic that deals with DFS paths, use that to do the actual open.
            open_at_end = True
            transaction = SMBFileTransaction(self)

        share_access = _parse_share_access(self.share_access, self.mode)
        create_disposition = _parse_mode(self.mode, invalid=self._INVALID_MODE)

        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wpo/feeb3122-cfe0-4b34-821d-e31c036d763c
        # Impersonation on SMB has little meaning when opening files but is important if using RPC so set to a sane
        # default of Impersonation.
        transaction += self.fd.create(
            ImpersonationLevel.Impersonation,
            self._desired_access,
            self._file_attributes,
            share_access,
            create_disposition,
            self._create_options,
            send=False,
        )

        if open_at_end:
            transaction.commit()
            if "a" in self.mode and self.FILE_TYPE != "pipe":
                self._offset = self.fd.end_of_file

    def readable(self):
        """True if file was opened in a read mode."""
        return "r" in self.mode or "+" in self.mode

    def seek(self, offset, whence=SEEK_SET):
        """
        Move to new file position and return the file position.

        Argument offset is a byte count.  Optional argument whence defaults to
        SEEK_SET or 0 (offset from start of file, offset should be >= 0); other values
        are SEEK_CUR or 1 (move relative to current position, positive or negative),
        and SEEK_END or 2 (move relative to end of file, usually negative, although
        many platforms allow seeking beyond the end of a file).

        Note that not all file objects are seekable.
        """
        seek_offset = {
            SEEK_SET: 0,
            SEEK_CUR: self._offset,
            SEEK_END: self.fd.end_of_file,
        }[whence]
        self._offset = seek_offset + offset
        return self._offset

    def seekable(self):
        """True if file supports random-access."""
        return True

    def tell(self):
        """
        Current file position.

        Can raise OSError for non seekable files.
        """
        return self._offset

    def truncate(self, size):
        """
        Truncate the file to at most size bytes and return the truncated size.

        Size defaults to the current file position, as returned by tell().
        The current file position is changed to the value of size.
        """
        with SMBFileTransaction(self) as transaction:
            eof_info = FileEndOfFileInformation()
            eof_info["end_of_file"] = size
            set_info(transaction, eof_info)

        self.fd.end_of_file = size
        self._flush = True
        return size

    def writable(self):
        """True if file was opened in a write mode."""
        return "w" in self.mode or "x" in self.mode or "a" in self.mode or "+" in self.mode

    def readall(self):
        """
        Read and return all the bytes from the stream until EOF, using
        multiple calls to the stream if necessary.

        :return: The byte string read from the SMB file.
        """
        data = bytearray()
        while True:
            read_length = min(
                # We always want to be reading a minimum of 64KiB.
                max(self.fd.end_of_file - self._offset, MAX_PAYLOAD_SIZE),
                self.fd.connection.max_read_size,  # We can never read more than this.
            )

            buffer = bytearray(b"\x00" * read_length)
            bytes_read = self.readinto(buffer)
            if not bytes_read:
                break

            data += buffer[:bytes_read]

        return bytes(data)

    def readinto(self, b):
        """
        Read bytes into a pre-allocated, writable bytes-like object b, and
        return the number of bytes read. This may read less bytes than
        requested as it depends on the negotiated read size and SMB credits
        available.

        :param b: bytes-like object to read the data into.
        :return: The number of bytes read.
        """
        if self._offset >= self.fd.end_of_file and self.FILE_TYPE != "pipe":
            return 0

        chunk_size, credit_request = _chunk_size(self.fd.connection, len(b), "read")

        log.debug("Read %d -%d.", self._offset, chunk_size)
        read_msg, recv_func = self.fd.read(self._offset, chunk_size, send=False)
        request = self.fd.connection.send(
            read_msg,
            sid=self.fd.tree_connect.session.session_id,
            tid=self.fd.tree_connect.tree_connect_id,
            credit_request=credit_request,
        )

        try:
            file_bytes = recv_func(request)
        except PipeBroken:
            # A pipe will block until it returns the data available or was closed/broken.
            file_bytes = b""

        b[: len(file_bytes)] = file_bytes

        if self.FILE_TYPE != "pipe":
            self._offset += len(file_bytes)

        return len(file_bytes)

    def write(self, b):
        """
        Write buffer b to file, return number of bytes written.

        Only makes one system call, so not all of the data may be written.
        The number of bytes actually written is returned. This can be less than
        the length of b as it depends on the underlying connection.
        """
        chunk_size, credit_request = _chunk_size(self.fd.connection, len(b), "write")
        data = bytes(b[:chunk_size])

        write_msg, recv_func = self.fd.write(data, offset=self._offset, send=False)
        request = self.fd.connection.send(
            write_msg,
            sid=self.fd.tree_connect.session.session_id,
            tid=self.fd.tree_connect.tree_connect_id,
            credit_request=credit_request,
        )
        bytes_written = recv_func(request)

        if self.FILE_TYPE != "pipe":
            self._offset += bytes_written
            self.fd.end_of_file = max(self.fd.end_of_file, self._offset)
            self._flush = True

        return bytes_written


class SMBDirectoryIO(SMBRawIO):
    FILE_TYPE = "dir"
    _INVALID_MODE = "w+"

    def query_directory(self, pattern, info_class):
        query_flags = QueryDirectoryFlags.SMB2_RESTART_SCANS
        while True:
            try:
                entries = self.fd.query_directory(pattern, info_class, flags=query_flags)
            except NoMoreFiles:
                break

            query_flags = 0  # Only the first request should have set SMB2_RESTART_SCANS
            yield from entries

    def readable(self):
        return False

    def seekable(self):
        return False

    def writable(self):
        return False


class SMBFileIO(SMBRawIO):
    FILE_TYPE = "file"


class SMBPipeIO(SMBRawIO):
    FILE_TYPE = "pipe"

    def seekable(self):
        return False
